#!/usr/bin/env python3
"""Wrapper for the CCSDS 122.1 (MHDC) codec
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/05/29"

import os
import math

import enb


class MHDC_Abstract(enb.icompression.LosslessCodec, enb.icompression.WrapperCodec):
    """Default values as in Green Book
    """

    def __init__(self, wavelet_2d_type="Integer", bin_dir=None, output_invocation_dir=None,
                 target_rate_bpppc=None):
        """
        :param wavelet_2d_type: "Integer" for lossless 2D DWT, "Float" for lossy 2D DWT
        :param target_rate_bpppc: if not None, it selects a target rate in bpppc for each iamge
        """
        bin_dir = bin_dir if bin_dir is not None else os.path.dirname(os.path.abspath(__file__))
        super().__init__(compressor_path=os.path.join(bin_dir, "mhdcEncoder.sh"),
                         decompressor_path=os.path.join(bin_dir, "mhdcDecoder.sh"),
                         output_invocation_dir=output_invocation_dir)
        assert wavelet_2d_type in ["Integer", "Float"]
        self.param_dict["wavelet_2d_type"] = wavelet_2d_type
        assert target_rate_bpppc is None or target_rate_bpppc > 0
        self.param_dict["target_rate_bpppc"] = round(target_rate_bpppc, 4) if target_rate_bpppc is not None else None

    def get_transform_params(self, original_file_info):
        """Return a string with the transformation params for each subclass.
        """
        raise NotImplementedError()

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        assert original_file_info["bytes_per_sample"] == 2, "CCSDS 122 configured for 16bit only"

        if self.param_dict["target_rate_bpppc"] is None:
            target_rate_param = ""
        else:
            target_rate_bytes = math.ceil(
                self.param_dict["target_rate_bpppc"] * original_file_info["samples"] / 8)
            target_rate_param = f"-r {target_rate_bytes} -a lagrange"

        return f"-i {original_path} " \
               f"-x {original_file_info['width']} " \
               f"-y {original_file_info['height']} " \
               f"-z {original_file_info['component_count']} " \
               f"-s {'yes' if original_file_info['signed'] else 'no'} " \
               f"-U {self.param_dict['U'] if 'U' in self.param_dict else 0} " \
               f"-D {self.param_dict['D'] if 'D' in self.param_dict else 0} " \
               f"-R {math.ceil(original_file_info['width'] / 8)} " \
               f"-S {self.param_dict['S'] if 'S' in self.param_dict else 128} " \
               f"-W {self.param_dict['W'] if 'W' in self.param_dict else 1} " \
               f"-w {self.param_dict['wavelet_2d_type']} " \
               f"-o {compressed_path} " \
               f"{target_rate_param} " \
               f"{self.get_transform_params(original_file_info=original_file_info)}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"-i '{compressed_path}' -o '{reconstructed_path}'"


class MHDC_ID(MHDC_Abstract):
    @property
    def label(self):
        return f"CCSDS 122.1"

    def get_transform_params(self, original_file_info):
        return f"-t id"


class MHDC_POT(MHDC_Abstract):
    def __init__(self, bin_dir=None, output_invocation_dir=None,
                 F=32, Omega=12, wavelet_2d_type="Integer",
                 target_rate_bpppc=None):
        super().__init__(bin_dir=bin_dir, output_invocation_dir=output_invocation_dir,
                         wavelet_2d_type=wavelet_2d_type, target_rate_bpppc=target_rate_bpppc)
        self.param_dict["F"] = F
        self.param_dict["Omega"] = Omega

    @property
    def label(self):
        return f"CCSDS 122.1 POT"

    def get_transform_params(self, original_file_info):
        return f"-t pot -F {self.param_dict['F']} -O {self.param_dict['Omega']}"


class MHDC_IWT(MHDC_Abstract):
    def __init__(self, bin_dir=None, output_invocation_dir=None,
                 wavelet_2d_type="Integer", target_rate_bpppc=None):
        """:param wavelet_2d_type: determines the type of DWT applied 
          in the spatial domain.
        """
        super().__init__(bin_dir=bin_dir, output_invocation_dir=output_invocation_dir,
                         wavelet_2d_type=wavelet_2d_type, target_rate_bpppc=target_rate_bpppc)

    @property
    def label(self):
        return f"CCSDS 122.1 IWT"

    def get_transform_params(self, original_file_info):
        return f"-t iwt"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Logging utilities for `enb`.

It uses only symbols from .misc, but no other module in enb.
"""
__author__ = "Miguel Hernández-Cabronero"
__date__ = "2021/08/13"

import contextlib
import sys
import time
import rich
import rich.progress
import rich.markup
import rich.console

from .misc import ExposedProperty
from .misc import Singleton
from . import config


class LogLevel:
    """Each of the available logging levels is an instance of this class. A
    level represents a named type of message, with a priority comparable to
    other levels.
    """

    # pylint: disable=too-few-public-methods,too-many-instance-attributes

    def __init__(self, name, priority=0, prefix=None, help_message=None, style=None):
        """
        :param priority: minimum priority level needed to show this level.
        :param name: unique name for the level.
        :param prefix: prefix when printing messages of this level. If None,
          a default one is used based on the name.
        :param help_message: optional help explaining the purpose of the level.
        :param style: if not None, a color with which messages of this level are displayed. See
          https://rich.readthedocs.io/en/stable/appendix/colors.html for more details about
          available colors. If None, the default color is used.
        """
        self.name = name
        self.priority = priority
        self.label = prefix
        self.help = help_message
        if prefix is not None:
            self.prefix = prefix
        else:
            self.prefix = f"[{name[0].upper()}] "
        self.style = style

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name}:{self.priority})"


@config.aini.managed_attributes
class Logger(metaclass=Singleton):
    """Message logging and printing hub for `enb`.

    Messages are only shown if their priority is at least as high as the
    configured minimum.

    The minimum required level name (from "core" down to "debug") can be
    selected via the CLI and the file-based configuration by setting the
    `selected_log_level` flag/option.

    You can then modify this minimum value programmatically by setting
    `enb.config.options.minimum_priority_level` to a new :class:`LogLevel`
    instance, such as LOG_VERBOSE or any of the other constants defined above.
    """

    # pylint: disable=too-many-instance-attributes,too-many-public-methods

    # Default style for core level messages
    style_core = "#28c9ff"
    # Default style for error level messages
    style_error = "bold #ff5255"
    # Default style for warning level messages
    style_warn = "#ffca4f"
    # Default style for message level messages
    style_message = "#28c9ff"
    # Default style for verbose level messages
    style_verbose = "#a5d3a5"
    # Default style for info level messages
    style_info = "#9b5ccb"
    # Default style for debug level messages
    style_debug = "#909090"
    # Style for the banner line
    banner_line_style = "#767676 bold"
    # Style for the regular text in the banner
    banner_plain_text_style = "#767676"
    # Style for the 'enb' part in the banner
    banner_enb_name_style = "#f3ac05 bold"
    # Style for the version part in the banner
    banner_enb_version_style = "#9b5ccb bold"

    def __init__(self):
        # Available logging levels and their intended usage
        self.levels = [
            LogLevel("core",
                     help_message="Messages always shown, "
                                  "no matter the priority level",
                     style=self.style_core),
            LogLevel("error",
                     help_message="A critical error that prevents "
                                  "from completing the main task",
                     style=self.style_error),
            LogLevel("warn",
                     help_message="Something wrong or bogus happened, "
                                  "but the main task can be completed",
                     style=self.style_warn),
            LogLevel("message",
                     help_message="Task-central messages intended "
                                  "to appear in console",
                     style=self.style_message),
            LogLevel("verbose",
                     help_message="Messages for the interested user, "
                                  "e.g., task progress",
                     style=self.style_verbose),
            LogLevel("info",
                     help_message="Messages for the very interested "
                                  "user/developer, e.g., detailed "
                                  "task progress",
                     style=self.style_info),
            LogLevel("debug",
                     help_message="Messages for debugging purposes, "
                                  "e.g., traces and watches",
                     style=self.style_debug),
        ]
        # Assign an integer priority level to each defined level (higher:
        # less priority).
        for i, level in enumerate(self.levels):
            level.priority = i
        # Keep easy access to the levels by their name
        self.name_to_level = {level.name: level for level in self.levels}

        # Minimum priority level required to be printed. It starts by the
        # least restrictive option, to allow logging during the configuration
        # load process without introducing circular references.
        self.selected_log_level = \
            sorted(self.name_to_level.values(),
                   key=lambda lev: lev.priority)[0]

        # Create a few IDE-friendly attributes that point to the predefined levels by name
        self.level_core = self.get_level("core")
        self.level_error = self.get_level("error")
        self.level_warn = self.get_level("warn")
        self.level_message = self.get_level("message")
        self.level_verbose = self.get_level("verbose")
        self.level_info = self.get_level("info")
        self.level_debug = self.get_level("debug")
        # This value is changed based on the file and CLI configuration by
        # enb/__init__.py
        self.show_prefixes = False
        # show_prefix_level determines the level required to include prefixes
        # in any shown messages.
        self.show_prefix_level = self.level_info

        # Last levels and end strings used when logging
        self._last_end = None
        self._last_level = None
        # Store the original builtins print function when replacing
        self._original_print = None
        # Lazy imports from parallel_ray, to avoid circular definitions
        self._is_parallel_process = None
        self._is_ray_enabled = None

    def levels_by_priority(self):
        """Return a list of the available levels, sorted from higher to lower
        priority.
        """
        return sorted(self.name_to_level.values(),
                      key=lambda level: level.priority)

    def log(self, msg, level, end="\n", file=None, markup=False, highlight=False, style=None,
            rule=False, rule_kwargs=None):
        """Conditionally log a message given its level. It only shares "end"
        with builtins.print as keyword argument.

        :param msg: message to be logged
        :param level: priority level for the message
        :param end: string appended after the message, if it is shown.
        :param file: file where to log the message, or None to automatically
          select sys.stdout
        :param markup: should rich markup be interpreted within the message?
        :param highlight: should rich apply automatic highlighting of numbers, constants, etc., to the message?
        :param style: if not None, the level's current style is overwritten by this
        :param rule: should the message be displayed with console.rule()?
        :param rule_kwargs: if rule_kwargs is True, these parameters are passed to console.rule
        """
        # pylint: disable=too-many-arguments
        file = file or sys.stdout

        if level.priority <= self.selected_log_level.priority:
            try:
                # pylint: disable=access-member-before-definition
                last_end = self._last_end
                last_level = self._last_level
            except AttributeError:
                last_end = "\n"
                last_level = self.selected_log_level

            forfeit_prefix = last_level is level and not last_end.endswith("\n")
            forfeit_prefix = forfeit_prefix \
                             or (self.selected_log_level.priority
                                 <= self.show_prefix_level.priority)
            split_message_str = "" if last_level is None \
                                      or last_level is level \
                                      or last_end is None \
                                      or last_end.endswith("\n") \
                else "\n"

            output_msg = \
                f"{split_message_str}" \
                + (str(level.prefix) if self.show_prefixes
                                        and not forfeit_prefix else '') + \
                f"{msg}{end}"

            try:
                from .progress import ProgressTracker
                console = ProgressTracker.console
                console.markup = markup
                console.highlight = highlight
            except AttributeError:
                console = None
            console = console or rich.console.Console(
                file=file, markup=markup, highlight=highlight,
                force_terminal=True, force_interactive=True, force_jupyter=False)
            style = style or level.style

            if rule:
                console.rule(output_msg, **(rule_kwargs or dict()))
            else:
                console.print(output_msg, end="", style=style, highlight=highlight, markup=markup)

            self._last_end = end
            self._last_level = level

    def show_banner(self, level=None):
        """Shows the enb banner, including the current version.

        :param level: the priority level with which the banner is shown. If None, verbose is used by default.
        """
        banner_contents = (
            f"[{self.banner_line_style}][bold])[/bold][/{self.banner_line_style}] "
            f"[{self.banner_plain_text_style}]"
            f"Powered by "
            f"[{self.banner_enb_name_style}]enb[/{self.banner_enb_name_style}] "
            f"[{self.banner_enb_version_style}]"
            f"v{config.ini.get_key('enb', 'version')}"
            f"[/{self.banner_enb_version_style}]"
            f"[/{self.banner_plain_text_style}]"
            f" [{self.banner_line_style}][bold]([/bold][/{self.banner_line_style}]"
        )

        level = level or self.level_verbose
        self.log("", level=level)
        self.log(banner_contents, end="", rule=True, markup=True,
                 rule_kwargs=dict(style=self.banner_line_style,
                                  align="center"), level=level)
        self.log("", level=level)

    @property
    def is_parallel_process(self):
        """Lazy property to determine whether this is currently a parallel process.
        """
        if self._is_parallel_process:
            is_parallel_process = self._is_parallel_process
        else:
            from .parallel_ray import is_parallel_process
            self._is_parallel_process = is_parallel_process
        return is_parallel_process

    @property
    def is_ray_enabled(self):
        """Lazy property to determine whether ray is available and enabled.
        """
        if self._is_ray_enabled:
            is_ray_enabled = self._is_ray_enabled
        else:
            from .parallel_ray import is_ray_enabled
            self._is_ray_enabled = is_ray_enabled
        return is_ray_enabled

    def core(self, msg, **kwargs):
        """A message of "core" level.

        :param kwargs: optional arguments passed to self.log (must be
          compatible)
        """
        self.log(msg=msg, level=self.level_core, **kwargs)

    def error(self, msg, **kwargs):
        """Log an error message.

        :param kwargs: optional arguments passed to self.log (must be
          compatible)
        """
        self.log(msg=msg, level=self.level_error, **kwargs)

    def warn(self, msg, **kwargs):
        """Log a warning message.

        :param kwargs: optional arguments passed to self.log (must be
          compatible)
        """
        self.log(msg=msg, level=self.level_warn, **kwargs)

    def message(self, msg, **kwargs):
        """Log a regular console message.

        :param kwargs: optional arguments passed to self.log (must be
          compatible)
        """
        self.log(msg=msg, level=self.level_message, **kwargs)

    def verbose(self, msg, **kwargs):
        """Log a verbose console message.

        :param kwargs: optional arguments passed to self.log (must be
          compatible)
        """
        self.log(msg=msg, level=self.level_verbose, **kwargs)

    def info(self, msg, **kwargs):
        """Log an extra-informative console message.

        :param kwargs: optional arguments passed to self.log (must be
          compatible)
        """
        self.log(msg=msg, level=self.level_info, **kwargs)

    def debug(self, msg, **kwargs):
        """Log a debug trace.

        :param kwargs: optional arguments passed to self.log (must be
          compatible)
        """
        self.log(msg=msg, level=self.level_debug, **kwargs)

    @contextlib.contextmanager
    def log_context(self, msg, level, sep="...", msg_after=None,
                    show_duration=True):
        """Log a message before executing the `with` block code, run the
        block, and log another message when the block is completed. The
        message given the selected priority level, and is only displayed
        based on `self.selected_log_level`. The block of code is executed
        regardless of the logging options.

        :param msg: Message typically describing the
        :param level: Priority level for the shown messages.
        :param sep: separator printed between msg_before and msg_after (
          newline is not required in it to allow single-line reporting)
        :param msg_after: message shown after `msg` and `sep` upon
          completion. If none, one is automatically selected based on msg.
        :param show_duration: if True, a message displaying the run time is
          logged upon completion.
        """
        # pylint: disable=too-many-arguments
        # Show entry message
        self.log(msg=msg, end=sep, level=level)
        time_before = time.time()

        # Run block
        yield None
        run_time = time.time() - time_before

        # Show exit message
        if msg_after is None:
            space = " "
            try:
                space = space if not self._last_end.endswith("\n") else ""
                msg_after = f"{space}done" if self._last_level is level and self._last_end == sep \
                    else f"done ({msg.rstrip()})"
            except AttributeError:
                msg_after = f"{space}done"
        if show_duration:
            msg_after += f" (took {run_time:.2f}s)"
        msg_after += "." if msg_after[-1] != "." else ""
        self.log(msg=msg_after, level=level)

    def core_context(self, msg, sep="...", msg_after=None, show_duration=True):
        """Logging context of core priority.

        :param msg: Message to show before starting the code block.
        :param sep: separator printed between msg_before and msg_after (
          newline is not required in it to allow single-line reporting).
        :param msg_after: message shown after `msg` and `sep` upon completion.
        :param show_duration: if True, a message displaying the run time is
          logged upon completion.
        """
        return self.log_context(msg=msg, level=self.level_core,
                                sep=sep, msg_after=msg_after,
                                show_duration=show_duration)

    def message_context(self, msg, sep="...", msg_after=None,
                        show_duration=True):
        """Logging context of message priority.

        :param msg: Message to show before starting the code block.
        :param sep: separator printed between msg_before and msg_after (
          newline is not required in it to allow single-line reporting)
        :param msg_after: message shown after `msg` and `sep` upon completion.
        :param show_duration: if True, a message displaying the run time is
          logged upon completion.
        """
        return self.log_context(msg=msg, level=self.level_message,
                                sep=sep, msg_after=msg_after,
                                show_duration=show_duration)

    def verbose_context(self, msg, sep="...", msg_after=None,
                        show_duration=True):
        """Logging context of verbose priority.

        :param msg: Message to show before starting the code block.
        :param sep: separator printed between msg_before and msg_after (
          newline is not required in it to allow single-line reporting)
        :param msg_after: message shown after `msg` and `sep` upon completion.
        :param show_duration: if True, a message displaying the run time is
          logged upon completion.
        """
        return self.log_context(msg=msg, level=self.level_verbose,
                                sep=sep, msg_after=msg_after,
                                show_duration=show_duration)

    def info_context(self, msg, sep="...", msg_after=None, show_duration=True):
        """Logging context of info priority.

        :param msg: Message to show before starting the code block.
        :param sep: separator printed between msg_before and msg_after (
          newline is not required in it to allow single-line reporting)
        :param msg_after: message shown after `msg` and `sep` upon completion.
        :param show_duration: if True, a message displaying the run time is
          logged upon completion.
        """
        return self.log_context(msg=msg, level=self.level_info,
                                sep=sep, msg_after=msg_after,
                                show_duration=show_duration)

    def debug_context(self, msg, sep="...", msg_after=None, show_duration=True):
        """Logging context of debug priority.

        :param msg: Message to show before starting the code block.
        :param sep: separator printed between msg_before and msg_after (
          newline is not required in it to allow single-line reporting)
        :param msg_after: message shown after `msg` and `sep` upon completion.
        :param show_duration: if True, a message displaying the run time is
          logged upon completion.
        """
        return self.log_context(msg=msg, level=self.level_debug,
                                sep=sep, msg_after=msg_after,
                                show_duration=show_duration)

    def level_active(self, name, **kwargs):
        """Return True if and only if the given name corresponds to a level
        with priority sufficient given self.min_priority_level.
        """
        # pylint: disable=unused-argument
        if isinstance(name, LogLevel):
            name = name.name
        return self.name_to_level[name].priority <= self.selected_log_level.priority

    @property
    def core_active(self):
        """Return True if and only if the core level is currently active,
        i.e., the current `self.min_priority_level` has a greater or equal
        priority value than the core level.
        """
        return self.level_active("core")

    @property
    def error_active(self):
        """Return True if and only if the error level is currently active,
        i.e., the current `self.min_priority_level` has a greater or equal
        priority value than the error level.
        """
        return self.level_active("error")

    @property
    def warn_active(self):
        """Return True if and only if the warn level is currently active,
        i.e., the current `self.min_priority_level` has a greater or equal
        priority value than the warn level.
        """
        return self.level_active("warn")

    @property
    def message_active(self):
        """Return True if and only if the message level is currently active,
        i.e., the current `self.min_priority_level` has a greater or equal
        priority value than the message level.
        """
        return self.level_active("message")

    @property
    def verbose_active(self):
        """Return True if and only if the verbose level is currently active,
        i.e., the current `self.min_priority_level` has a greater or equal
        priority value than the verbose level.
        """
        return self.level_active("verbose")

    @property
    def info_active(self):
        """Return True if and only if the info level is currently active,
        i.e., the current `self.min_priority_level` has a greater or equal
        priority value than the info level.
        """
        return self.level_active("info")

    @property
    def debug_active(self):
        """Return True if and only if the debug level is currently active,
        i.e., the current `self.min_priority_level` has a greater or equal
        priority value than the debug level.
        """
        return self.level_active("debug")

    def report_level_status(self):
        """:return: a string reporting the present logging levels and whether
          or not they are active.
        """
        lines = [f"{'level':8s}  {'priority':8s}  {'active':6s}"]
        lines.append("-" * len(lines[0]))
        lines.extend(
            f"{name:8s}  {str(level.priority):8s}  {self.level_active(name)}"
            for name, level in self.name_to_level.items())
        return "\n".join(lines)

    def get_level(self, name, lower_priority=0):
        """If lower_priority is 0, return the logging level associated with
        the name passed as argument. Otherwise, the aforementioned level's
        priority is lowered by that numeric amount (positive values means
        less prioritary levels can be selected).

        After that, the available level with the closest priority is chosen.
        """
        # Obtain the logger for the given name
        base_level = self.name_to_level[name]

        # Shift priority if requested - always return something in the list
        # of available levels
        if lower_priority != 0:
            new_priority = base_level.priority + int(lower_priority)
            levels_by_priority = logger.levels_by_priority()
            base_level = levels_by_priority[0]
            for level in levels_by_priority[1:]:
                if level.priority <= new_priority:
                    base_level = level
                else:
                    break

        return base_level

    def print_to_log(self, *args, sep=" ", end="\n", file=None):
        """Method used to substitute print if configured to do so.
        If file is None, then sys.stdout is used by default.
        """
        self.message(f"{sep.join((str(a) for a in args))}", end=end, file=file)

    def __repr__(self):
        return f"{self.__class__.__name__}(selected={self.selected_log_level})"


# Singleton instance of the logger, shared across modules even if reinstantiated.
logger = Logger()
assert logger is Logger(), "Singleton not working for log.py"

# Expose logging functions
get_level = logger.get_level
log = logger.log
core = logger.core
error = logger.error
warn = logger.warn
message = logger.message
verbose = logger.verbose
info = logger.info
debug = logger.debug

# Expose functions to check whether a level is active or not
core_active = ExposedProperty(instance=logger, property_name="core_active")
error_active = ExposedProperty(instance=logger, property_name="error_active")
warn_active = ExposedProperty(instance=logger, property_name="warn_active")
message_active = ExposedProperty(instance=logger, property_name="message_active")
verbose_active = ExposedProperty(instance=logger, property_name="verbose_active")
info_active = ExposedProperty(instance=logger, property_name="info_active")
debug_active = ExposedProperty(instance=logger, property_name="debug_active")

# Expose status report functions
show_banner = logger.show_banner
report_level_status = logger.report_level_status



import enb.plugins


class FLIFPlugin(enb.plugins.PluginMake):
    name = "flif"
    label = "Wrapper for the FLIF compressor"
    tags = {"data compression", "codec"}
    contrib_authors = ["The FLIF community"]
    contrib_reference_urls = ["https://flif.info"]
    contrib_download_url_name = [
        ("https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/flif_0.3_2017.zip?raw=true",
         "flif_0.3_2017.zip")]
    extra_requirements_message = """
    Additional packages are needed for this plugin. These need to be installed via pip, pacman, or 
    whatever system package manager you use. 
    
    Instructions for several linux distributions are shown next, copied from 
    the README.md provided in the original package:
        
      * Debian: `sudo apt update; sudo apt-get install libpng-dev`
      * Ubuntu: `sudo apt update; sudo apt install -y libpng-dev make pgk-config`
      * Fedora: `sudo apt update; sudo dnf install libpng-devel`
    
    Instructions for Windows, MacOS and others can also be found in the README.md file.        
    """
    tested_on = {"linux"}



"""Codecs implement the `compress` and `decompress` methods
as well as a `name` and `label` to identify and represent them.

A `param_dict` is passed on initialization that describes the configuration of each codec instance.
Codecs may choose the number of parameters and their names.  
"""
import shutil
from enb.experiment import ExperimentTask
from enb.compression import CompressionResults, DecompressionResults

class AbstractCodec(ExperimentTask):
    """Base class for all codecs.
    """

    def __init__(self, param_dict=None):
        """
        :param param_dict: dictionary of parameters for this codec instance.
        """
        super().__init__(param_dict=param_dict)

    @property
    def name(self):
        """Name of the codec. Subclasses are expected to yield different
        values when different parameters are used. By default, the class name
        is folled by all elements in self.param_dict sorted alphabetically
        are included in the name.
        """
        name = f"{self.__class__.__name__}"
        if self.param_dict:
            name += "__" + "_".join(
                f"{k}={v}" for k, v in sorted(self.param_dict.items()))
        return name

    @property
    def label(self):
        """Label to be displayed for the codec. May not be strictly unique
        nor fully informative. By default, self's class name is returned.
        """
        return self.__class__.__name__

    def compress(self, original_path: str, compressed_path: str, original_file_info=None) -> CompressionResults:
        """Compress original_path into compress_path using param_dict as params.
        :param original_path: path to the original file to be compressed
        :param compressed_path: path to the compressed file to be created
        :param original_file_info: a dict-like object describing
          original_path's properties (e.g., geometry), or None.
        :return: (optional) a CompressionResults instance, or None
          (see self.compression_results_from_paths)
        """
        raise NotImplementedError()

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        """Decompress compressed_path into reconstructed_path using param_dict
        as params (if needed).

        :param compressed_path: path to the input compressed file
        :param reconstructed_path: path to the output reconstructed file
        :param original_file_info: a dict-like object describing
          original_path's properties (e.g., geometry), or None. Should only be
          actually used in special cases, since codecs are expected to store
          all needed metainformation in the compressed file.
        :return: (optional) a DecompressionResults instance, or None (see
        self.decompression_results_from_paths)
        """
        raise NotImplementedError()

    def compression_results_from_paths(self, original_path, compressed_path):
        """Get the default CompressionResults instance corresponding to
        the compression of original_path into compressed_path
        """
        return CompressionResults(
            codec_name=self.name,
            codec_param_dict=self.param_dict,
            original_path=original_path,
            compressed_path=compressed_path,
            compression_time_seconds=None)

    def decompression_results_from_paths(
            self, compressed_path, reconstructed_path):
        """Return a enb.icompression.DecompressionResults instance given
        the compressed and reconstructed paths.
        """
        return DecompressionResults(
            codec_name=self.name,
            codec_param_dict=self.param_dict,
            compressed_path=compressed_path,
            reconstructed_path=reconstructed_path,
            decompression_time_seconds=None)

    def __repr__(self):
        return f"<{self.__class__.__name__}(" \
            + ', '.join(repr(param) + '=' + repr(value)
                        for param, value in self.param_dict.items()) \
            + ")>"

class LosslessCodec(AbstractCodec):  # pylint: disable=abstract-method
    """An AbstractCodec that identifies itself as lossless.
    """


class LossyCodec(AbstractCodec):  # pylint: disable=abstract-method
    """An AbstractCodec that identifies itself as lossy.
    """


class NearLosslessCodec(LossyCodec):  # pylint: disable=abstract-method
    """An AbstractCodec that identifies itself as near lossless.
    """
    
class PassthroughCodec(LosslessCodec):
    """Codec that simply copies the input into the output in both compression and decompression.
    """
    def __init__(self):
        super().__init__()
    
    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        shutil.copyfile(original_path, compressed_path)
        
    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        shutil.copyfile(compressed_path, reconstructed_path)


import os
import zipfile
import shutil
import enb.plugins


class RLE(enb.plugins.PluginMake):
    name = "rle"
    label = "Run-Length Encoding codec"
    tags = {"data compression", "codec"}
    contrib_authors = ["Michael Dipperstein"]
    contrib_reference_urls = ["https://github.com/MichaelDipperstein/rle"]
    tested_on = {"linux"}

    @classmethod
    def build(cls, installation_dir):
        with zipfile.ZipFile(os.path.join(os.path.dirname(__file__), "rle.zip"), "r") as zip_file:
            zip_file.extractall(path=installation_dir)
        super().build(installation_dir=installation_dir)
        shutil.rmtree(os.path.join(installation_dir, "rle_src"))



import enb


class MontecarloPiTemplate(enb.plugins.Template):
    """Experiment that shows how to approximate pi in a distributed way
    (for demonstration purposes only).
    """
    name = "montecarlo-pi"
    author = ["Miguel Hernández-Cabronero"]
    label = "Demo project that approximates pi in a distributed way"
    tags = {"documentation"}
    tested_on = {"linux", "windows"}

    @classmethod
    def build(cls, installation_dir):
        super().build(installation_dir=installation_dir)
        enb.plugins.get_installable_by_name("cluster-config").install(installation_dir=installation_dir)



import enb


class BasicWorkflowTemplate(enb.plugins.Template):
    """Template for lossless and lossy data compression experiments.
    """
    name = "basic-workflow"
    author = ["Miguel Hernández-Cabronero"]
    label = "Basic, self-contained example of enb's workflow"
    tags = {"project", "data compression", "documentation"}
    tested_on = {"linux", "windows"}


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Template and use example for the enb.sets.FileVersionTable class.

Takes a set of *.txt files and performs a simplistic normalization.
"""
__author__ = "Miguel Hernández Cabronero <miguel.hernandez@uab.cat>"
__date__ = "26/10/2022"

import enb


class TextNormalizationTable(enb.sets.FileVersionTable):
    dataset_files_extension = "txt"

    def version(self, input_path, output_path, row):
        with open(input_path, "r") as input_file, open(output_path, "w") as output_file:
            contents = input_file.read()
            output_file.write("\n".join(l.lower().strip() for l in contents.splitlines()))


if __name__ == '__main__':
    tnt = TextNormalizationTable(original_base_dir="original_data", version_base_dir="versioned_data",
                                 csv_support_path="")
    tnt.get_df()


#!/usr/bin/env python3
"""Unit tests for isets.py
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/04/07"

import unittest
import tempfile
import os
import numpy as np

import test_all
from enb import isets


class TestiSets(unittest.TestCase):

    def test_array_load_dump(self):
        """Test that loading and dumping of numpy arrays
        indexed by [x,y,z] is correctly performed.
        """
        width = 4
        height = 2
        component_count = 3

        for order in ["bsq", "bip", "bil"]:
            for is_float in [True, False]:
                bytes_per_sample_values = [2, 4, 8] if is_float else [1, 2, 4, 8]
                for bytes_per_sample in bytes_per_sample_values:
                    for signed in [True, False]:
                        value_factors = [1] if not signed else [1, -1]
                        for value_factor in value_factors:
                            big_endian_values = [True] if (bytes_per_sample == 1 or is_float) \
                                else [True, False]
                            for big_endian in big_endian_values:
                                self.run_one_case(big_endian, bytes_per_sample,
                                                  component_count, height,
                                                  is_float, order, signed,
                                                  value_factor, width)

    def run_one_case(self, big_endian, bytes_per_sample, component_count,
                     height, is_float, order, signed, value_factor, width):
        row = dict(
            width=width, height=height, component_count=component_count,
            float=is_float,
            signed=signed, bytes_per_sample=bytes_per_sample,
            big_endian=big_endian)
        array = np.zeros((width, height, component_count),
                         dtype=isets.iproperties_row_to_numpy_dtype(row))
        i = 0
        for z in range(component_count):
            for y in range(height):
                for x in range(width):
                    array[x, y, z] = i * value_factor
                    i += 1
        fd, temp_file_path = tempfile.mkstemp(
            suffix=f"_{isets.iproperties_row_to_sample_type_tag(row)}.raw")
        try:
            dtype_string = isets.iproperties_row_to_numpy_dtype(row)
            isets.dump_array(array, temp_file_path, mode="wb",
                             dtype=dtype_string, order=order)
            loaded_array = isets.load_array(
                file_or_path=temp_file_path,
                image_properties_row=row,
                order=order)
            assert (array == loaded_array).all(), (array, loaded_array)
            assert os.path.getsize(temp_file_path) \
                   == width * height * component_count * bytes_per_sample
        finally:
            try:
                os.remove(temp_file_path)
                os.close(fd)
            except OSError:
                pass


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""Test the tarlite module
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/04/08"

import os
import glob
import tempfile
import filecmp
import unittest

__file__ = os.path.abspath(__file__)
import test_all

import enb
from enb import tarlite


class TestTarlite(unittest.TestCase):
    def test_read_write(self):
        with tempfile.NamedTemporaryFile() as tmp_tarlite_file:
            input_paths = [p for p in glob.glob(os.path.join(os.path.abspath(os.path.dirname(__file__)), "*.py"))
                           if os.path.isfile(p)]
            tw = tarlite.TarliteWriter(initial_input_paths=input_paths)
            tw.write(output_path=tmp_tarlite_file.name)

            tr = tarlite.TarliteReader(tarlite_path=tmp_tarlite_file.name)
            with tempfile.TemporaryDirectory() as tmp_extract_dir:
                tr.extract_all(output_dir_path=tmp_extract_dir)

                for input_path in input_paths:
                    check_path = os.path.join(tmp_extract_dir, os.path.basename(input_path))
                    assert filecmp.cmp(input_path, check_path)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""Unit tests for experiment.py
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2023/02/17"

import unittest
import enb


class TestExperimentTask(unittest.TestCase):
    def test_task_equality(self):
        assert enb.experiment.ExperimentTask() \
               == enb.experiment.ExperimentTask()
        assert enb.experiment.ExperimentTask() \
               == enb.experiment.ExperimentTask(dict())
        assert enb.experiment.ExperimentTask(dict()) \
               == enb.experiment.ExperimentTask()

        assert enb.experiment.ExperimentTask(dict(a=1, b=2)) \
               == enb.experiment.ExperimentTask(dict(a=1, b=2))
        assert enb.experiment.ExperimentTask(dict(a=1, b=2)) \
               != enb.experiment.ExperimentTask(dict(a=1, b=2, c=3))
        assert enb.experiment.ExperimentTask(dict(a=1, b=2)) \
               != enb.experiment.ExperimentTask(dict(a=1))
        assert enb.experiment.ExperimentTask(dict(a=1, b=2)) \
               != enb.experiment.ExperimentTask(dict(a=1, b=100))
        assert enb.experiment.ExperimentTask(dict(a=1, b=2)) \
               != enb.experiment.ExperimentTask(dict(a=1))
        assert enb.experiment.ExperimentTask(dict(a=1, b=2)) \
               != enb.experiment.ExperimentTask(dict())

        assert enb.experiment.ExperimentTask(dict(a=1, b=2)) == \
               enb.experiment.ExperimentTask(dict(a=1, b=2))
        assert enb.experiment.ExperimentTask(dict(a=1, b=2, c=3)) \
               != enb.experiment.ExperimentTask(dict(a=1, b=2))
        assert enb.experiment.ExperimentTask(dict(a=1)) \
               != enb.experiment.ExperimentTask(dict(a=1, b=2))
        assert enb.experiment.ExperimentTask(dict(a=1, b=100)) \
               != enb.experiment.ExperimentTask(dict(a=1, b=2))
        assert enb.experiment.ExperimentTask(dict(a=1)) \
               != enb.experiment.ExperimentTask(dict(a=1, b=2))
        assert enb.experiment.ExperimentTask(dict()) \
               != enb.experiment.ExperimentTask(dict(a=1, b=2))

        class OtherTask(enb.experiment.ExperimentTask):
            pass

        assert enb.experiment.ExperimentTask() != OtherTask()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""Curate a dir of PPM files into another dir with their raw BSQ equivalents, 
using names that reflect their geometry.  
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2025/04/23"

import argparse
import os
import sys

import enb
from enb.compression.pgm import PPMCurationTable


def ppm_to_raw(ppm_dir_path: str | os.PathLike, raw_dir_path: str | os.PathLike):
    PPMCurationTable(original_base_dir=ppm_dir_path,
                     version_base_dir=raw_dir_path,
                     csv_support_path=None).get_df()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='PPM to raw BSQ images')
    parser.add_argument("ppm_dir", type=str, help="Directory with the PPM samples. Must exist.")
    parser.add_argument("raw_dir", type=str, help="Directory where the raw BSQ images are stored.")
    opts = parser.parse_args()

    if not os.path.isdir(opts.ppm_dir):
        enb.logger.error(f"PPM dir {opts.ppm_dir} does not exist.")
        sys.exit(1)
    if os.path.exists(opts.raw_dir):
        enb.logger.error(f"Raw BSQ dir {opts.raw_dir} already exists. Refusing to overwrite.")
        sys.exit(1)

    ppm_to_raw(ppm_dir_path=opts.ppm_dir, raw_dir_path=opts.raw_dir)



#!/usr/bin/env python3
"""Wrappers for the LC framework implementation
"""
__author__ = "Pau Quintas Torra, Xavier Fernandez Mellado"
__since__ = "2024/04/03"

import os
import re
import tempfile
import shutil

from enb import icompression
import enb


class LosslessLCFrameworkExperiment(enb.icompression.LosslessCompressionExperiment):
    """Lossless compression experiment with a `best_pipeline` column with the name
    of the best compressor for the current sample.
    """

    @enb.atable.column_function([
        enb.atable.ColumnProperties("best_pipeline", label="Best Compression Pipeline")
    ])
    def best_pipeline(self, index, row):
        """Name of the best pipeline for the current sample, e.g., 'BIT_2 RRE_1'
        """
        original_path, codec = self.index_to_path_task(index)
        if isinstance(codec, LCFramework) and original_path in codec.path_to_best_pipeline:
            row["best_pipeline"] = codec.path_to_best_pipeline[original_path]
        else:
            row["best_pipeline"] = "-"

    @enb.atable.column_function([
        enb.atable.ColumnProperties("generation_time_seconds", label="Generation time (s)"),
        enb.atable.ColumnProperties("compression_time_no_generation", label="Compression time (s)"),
    ])
    def set_time_measurements(self, index, row):
        """Columns related to the measurement of the LC search and generation times:

        - generation_time_seconds: time spend searching, generating and compiling the optimal codec.
        - compression_time_no_generation: compression time, not taking into account any time spent
          generating the optimal codec.
        """
        original_path, codec = self.index_to_path_task(index)
        if isinstance(codec, LCFramework) and original_path in codec.path_to_generation_time:
            row["generation_time_seconds"] = codec.path_to_generation_time[original_path]
        else:
            row["generation_time_seconds"] = 0

        row["compression_time_no_generation"] = (
                row["compression_time_seconds"] - row["generation_time_seconds"])



class LCFramework(icompression.WrapperCodec, icompression.LosslessCodec):
    """Codec based on the LC Framework. The `lc` binary is first used to search and generate
    the best compressor pipeline. This best compressor is compiled and executed on each
    call, and the time is included in the final result."""

    def __init__(self, compressor_path=__file__, decompressor_path=__file__,
                 combination=".+ .+"):
        """
        :param compressor_path: ignored and set automatically - kept for signature compatibility.
        :param decompressor_path: ignored and set automatically - kept for signature compatibility.
        :param combination: stage definition, e.g., ".+ .+" for the best combination of any two
          LC stage components.
        """
        super().__init__(compressor_path=compressor_path,
                         decompressor_path=decompressor_path,
                         param_dict=dict(combination=combination))

        self.tmp_dir = tempfile.mkdtemp(dir=enb.config.options.base_tmp_dir)
        self.combination_len = len(combination.split())
        # Name of the best pipeline for each sample indexed by the sample's path
        self.path_to_best_pipeline = {}
        # Generation time (including execution of lc and compilation of the codec), in seconds,
        # indexed by the sample's path
        self.path_to_generation_time = {}

    def compress(self, original_path, compressed_path, original_file_info=None):
        """Run lc to get the best stage combination, generate the code, compile, and compress
        with the compressor using that best combination.
        """
        execution_dir = os.path.join(self.tmp_dir, os.path.basename(compressed_path))
        self.path_to_generation_time[original_path] = 0

        # Run lc to find the best combination
        invocation = (f"{os.path.join(os.path.dirname(os.path.abspath(__file__)), 'lc')} "
                      f"{original_path} CR \"\" \"{self.param_dict['combination']}\"")
        enb.logger.debug(f"[{self.name}] Invocation: '{invocation}'")
        try:
            enb.logger.debug(f"[{self.name}] executing: {repr(invocation)}")
            status, output, measured_time, memory_kb = \
                enb.tcall.get_status_output_time_memory(invocation=invocation)
            self.path_to_generation_time[original_path] += measured_time
            enb.logger.debug(
                f"[{self.name}] Compression OK; "
                f"invocation={invocation} - status={status}; "
                f"output={output}; memory={memory_kb} KB")
            os.remove(os.path.join(os.getcwd(), os.path.basename(
                original_path) + f".CR{self.combination_len}.csv"))
        except enb.tcall.InvocationError as ex:
            try:
                os.remove(os.path.join(os.getcwd(), os.path.basename(
                    original_path) + f".CR{self.combination_len}.csv"))
            except:
                pass

            raise enb.icompression.CompressionException(
                original_path=original_path,
                compressed_path=compressed_path,
                file_info=original_file_info,
                status=-1,
                output=None) from ex

        # Compile compressor and decompressor with the best compression
        shutil.copytree(
            os.path.join(os.path.dirname(os.path.abspath(__file__)), "LC-framework-main"),
            os.path.join(execution_dir, "LC"))

        best_pipeline = self.output_to_best_pipeline(output)
        self.path_to_best_pipeline[original_path] = best_pipeline

        try:
            # Generate the compressor and decompressor source code
            invocation = (
                    os.path.join(execution_dir,
                                 'LC/generate_standalone_CPU_compressor_decompressor.py')
                    + f' "" "{best_pipeline}"'
            )
            _, _, measured_time, _ = \
                enb.tcall.get_status_output_time_memory(invocation=invocation)
            self.path_to_generation_time[original_path] += measured_time

            # Compile the compressor
            invocation = (
                f"g++ -O3 -march=native -mno-fma -I. -std=c++17 "
                f"-o {os.path.join(execution_dir, 'compress')} "
                f"{os.path.join(execution_dir, 'LC/compressor-standalone.cpp')}"
            )
            _, _, measured_time, _ = \
                enb.tcall.get_status_output_time_memory(invocation=invocation)
            self.path_to_generation_time[original_path] += measured_time

            # Compile the decompressor
            invocation = (
                f"g++ -O3 -march=native -mno-fma -I. -std=c++17 -o {os.path.join(execution_dir, 'decompress')} "
                f"{os.path.join(execution_dir, 'LC/decompressor-standalone.cpp')}"
            )
            _, _, measured_time, _ = \
                enb.tcall.get_status_output_time_memory(invocation=invocation)
            self.path_to_generation_time[original_path] += measured_time
        except enb.tcall.InvocationError as ex:
            raise enb.icompression.CompressionException(
                original_path=original_path,
                compressed_path=compressed_path,
                file_info=original_file_info,
                status=-1,
                output=None) from ex

        # Call to compressor using the best configuration
        self.compressor_path = os.path.join(execution_dir, "compress")
        self.decompressor_path = os.path.join(execution_dir, "decompress")
        result = super().compress(original_path, compressed_path, original_file_info)

        # The generation and compilation times are added to the total compression time
        result.compression_time_seconds += self.path_to_generation_time[original_path]

        return result

    def decompress(self, compressed_path, reconstructed_path, original_file_info):
        super().decompress(compressed_path=compressed_path,
                           reconstructed_path=reconstructed_path,
                           original_file_info=original_file_info)

        # Cleanup the temporary generation/compilation dir used for this compress/decompress
        if os.path.abspath(os.path.dirname(self.compressor_path)).startswith(
                os.path.abspath(self.tmp_dir)):
            shutil.rmtree(self.compressor_path, ignore_errors=True)

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        return f"{original_path} {compressed_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"{compressed_path} {reconstructed_path}"

    def output_to_best_pipeline(self, output: str) -> str:
        """Parse the output of an LC invocation and return the name of the best pipeline.
        """
        pattern = r'pipeline:\s*(.*?)\n'
        return re.findall(pattern, output)[-1]

    @property
    def label(self):
        return f"LC ({self.param_dict['combination']})"

    def __delete__(self, instance):
        """The codec is responsible for cleaning up its temporary directory.
        """
        shutil.rmtree(self.tmp_dir, ignore_errors=True)



#!/usr/bin/env python3
"""Wrappers for the reference JPEG implementation
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2023/01/12"

import os
from enb import icompression


class LPAQ8(icompression.WrapperCodec, icompression.LosslessCodec):
    def __init__(self, N=9,
                 compressor_path=os.path.join(os.path.dirname(os.path.abspath(__file__)), "lpaq8"),
                 decompressor_path=os.path.join(os.path.dirname(os.path.abspath(__file__)), "lpaq8"),
                 output_invocation_dir=None,
                 signature_in_name=False):
        """
        :param N: Memory usage is 6 + 3*2^N MB  (9 to 1542 MB). 
          Larger numbers usually give better compression at similar speed.
        """
        assert N == int(N), f"N must be an integer value"
        assert 0 <= N <= 9, f"N must be between 0 and 9, both included"
        super().__init__(compressor_path=compressor_path,
                         decompressor_path=decompressor_path,
                         param_dict=dict(N=int(N)),
                         output_invocation_dir=output_invocation_dir,
                         signature_in_name=signature_in_name)

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        return f"{self.param_dict['N']} {original_path} {compressed_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"d {compressed_path} {reconstructed_path}"

    @property
    def label(self):
        return f"LPAQ8 N{self.param_dict['N']}"



#!/usr/bin/env python3
"""Codec wrapper for the LZ4 lossless image coder
"""
__author__ = "Òscar Maireles and Miguel Hernández-Cabronero"
__since__ = "2021/06/01"

import os
import enb


class Lz4(enb.icompression.LosslessCodec, enb.icompression.NearLosslessCodec, enb.icompression.WrapperCodec):
    """Wrapper for the LZ4 codec
    All data types integer and float 16, 32, 64 can be compressed 
    """

    def __init__(self, lz4_binary=os.path.join(os.path.dirname(__file__), "lz4"), compression_level=9):
        """
        :param compression_level: 1-9, being 9 the maximum data reduction
        """
        super().__init__(compressor_path=lz4_binary,
                         decompressor_path=lz4_binary,
                         param_dict=dict(compression_level=compression_level))

    @property
    def label(self):
        return f"LZ4 level {self.param_dict['compression_level']}"

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        assert original_file_info["big_endian"] and not original_file_info["float"], \
            f"Only big-endian integer samples are currently supported by {self.__class__.__name__}"
        return f"-{self.param_dict['compression_level']} {original_path}  {compressed_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"-d -{self.param_dict['compression_level']} {compressed_path} {reconstructed_path} "



import os
import glob
import shutil
import configparser
import enb
from enb.config import options


class AbstractColorsTemplate:
    author = ["Miguel Hernández-Cabronero"]
    tags = {"project"}
    tested_on = {"linux"}

    @classmethod
    def install(cls, installation_dir):
        input_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), f"{cls.name}.ini")
        if not os.path.exists(input_path):
            raise ValueError(f"Template plugin {cls.__name__} is not properly configured: "
                             f"{input_path} does not exist.\nPlease reinstall enb and contact the "
                             f"maintainers if the problem persists.")

        output_path = os.path.join(installation_dir, f"{cls.name}.ini")
        if os.path.exists(output_path) and not options.force:
            raise ValueError(f"Output file {output_path} exists. Refusing to overwrite.")

        # Verify that no clashes exist with currently installed config files
        color_config = configparser.ConfigParser()
        color_config.read(input_path)

        for ini_path in glob.glob(os.path.join(installation_dir, "*.ini")):
            existing_config = configparser.ConfigParser()
            existing_config.read(ini_path)

            for cat_name, category in color_config.items():
                if cat_name == "DEFAULT":
                    continue
                if cat_name not in existing_config:
                    continue
                for color_option_name in category.keys():
                    if color_option_name in existing_config[cat_name]:
                        enb.logger.warn(
                            f"Option {repr(color_option_name)} in category {repr(cat_name)} "
                            f"was already present in {repr(os.path.abspath(ini_path))}, "
                            f"and in {repr(os.path.abspath(output_path))}. "
                            f"Consider removing one of the definitions, as it is not guaranteed"
                            f"which of the values will be kept by enb.\n",
                            markup=True)

        shutil.copyfile(input_path, output_path)
        cls.report_successful_installation(installation_dir=installation_dir)


class DefaultColorsTemplate(AbstractColorsTemplate, enb.plugins.Template):
    """Default color template.
    """
    name = "colors-default"
    label = "Default color template for light or dark background terminals"


class DarkColorsTemplate(AbstractColorsTemplate, enb.plugins.Template):
    """Dark color template.
    """
    name = "colors-dark"
    label = "Color template for dark background terminals"



#!/usr/bin/env python3
"""Tools define plugin installables.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/08/01"

import os
import glob
import inspect
import shutil
import platform
import subprocess
import textwrap
import enb

from .installable import Installable


class Plugin(Installable):
    """Plugins are self-contained, python modules that may assume enb is installed.

    - They can be installed into your projects via the enb CLI (e.g., with `enb plugin install <name> <clone_dir>`),
      and then imported like any other module.

    - The list of plugins available off-the-box can be obtained using the enb CLI (e.g., with `enb plugin list`).

    - Plugins may declare pip dependencies, which are attempted to be satisfied automatically when the plugins
      are installed. In addition, plugins may define their `extra_requirements_message` member to be not None,
      in which case it describes manual intervention required from the user either as a pre-installation
      or post-installation step. That message is shown when attempting to install the plugin with
      not-None `extra_requirements_message`.

    - The __init__.py file is preserved if present. If not present, one is created automatically,
      which imports all symbols from all .py modules in the plugin.
    """

    @classmethod
    def install(cls, installation_dir, overwrite_destination=False):
        """Make a copy of this plugin into `installation_dir`, ready to be imported.
        By default, a verbatim copy of the source plugin's dir is made.
        Any previous contents in installation_dir are overwritten.
        Then any explicit requirements are met (external software may be downloaded
        and pip packages installed).

        :param installation_dir: destination dir where the plugin is to be copied to and, when necessary, built.
        :param overwrite_destination: if True, the destination path is deleted before
          installation. If False and installation_dir already exists, an error
          is raised (plugins are intended to be self-contained, isolated python modules).
        """
        installation_dir_existed = os.path.exists(installation_dir)
        if not overwrite_destination and installation_dir_existed:
            raise ValueError(f"Plugin {repr(cls)} cannot be installed into existing "
                             f"path {installation_dir} because "
                             f"overwrite_destination={overwrite_destination}.")
        try:
            super().install(installation_dir=installation_dir,
                            overwrite_destination=True)
        except Exception as ex:
            print(f"Error installing plugin {repr(cls.name)}!\n")
            raise ex

        if enb.config.options.verbose:
            cls.print_info()

    @classmethod
    def build(cls, installation_dir):
        """Perform any additional retrieval, compilation and setup necessary for this plugin
        to be importable and usable. By default:

        - The existence of installation_dir as a directory is performed.
        - Install any required apt modules
        - Any needed python modules are installed via pip
        - Cleanup any generic files that might not be needed at this point
        - The __init__.py file is preserved or generated automatically
        """
        # Existence assertion
        assert os.path.isdir(installation_dir), \
            (f"{cls.__name__}.build(installation_dir={repr(installation_dir)}): "
             f"installation_dir does not exist and it could not be automatically created. "
             f"Verify you have write permissions on the parent folder(s).")

        # cleanup before installation
        shutil.rmtree(os.path.join(installation_dir, "__pycache__"), ignore_errors=True)

        # add custom __init__ if needed
        init_path = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(cls))),
                                 "__init__.py")
        with open(os.path.join(installation_dir, "__init__.py"), "w") as init_file:
            if os.path.exists(init_path):
                # An __init__.py file already existed, simply copy it
                with open(init_path, "r") as source_init:
                    init_file.write(source_init.read())
            else:
                # No __init__.py found. Generating one that publishes all symbols by default
                for py_path in [p for p in glob.glob(os.path.join(installation_dir, "*.py"))
                                if not os.path.basename(p).startswith("__")]:
                    module_name = os.path.basename(py_path)[:-3]
                    init_file.write(f"from . import {module_name}\n")
                    init_file.write(f"from .{module_name} import *\n\n")


class PluginMake(Plugin):
    """Plugin that assumes the existence of a valid Makefile in the installation folder,
    and uses it for building the plugin.
    """

    @classmethod
    def build(cls, installation_dir):
        platform_name = platform.system().lower()

        if shutil.which("make") is None:
            msg = (f"The {repr(cls.name)} plugin requires the `make` program to be "
                   f"installed and available in the path so that it can be installed. "
                   f"However, `make` was not found. ")
            if platform_name == "linux":
                msg += ("Please try installing it with `apt install build-essential` "
                        "or the equivalent for your distribution.")
            elif platform_name == "darwin":
                msg += ("You might be able to install it with `xcode-select --install` "
                        "or an alternative way for your platform.")
            elif platform_name == "windows":
                msg += ("You will need to install it manually. You might want to try "
                        "`winget install GnuWin32.Make` or some tools such as "
                        "https://chocolatey.org/ might be of help. Additional information "
                        "may be available at https://stackoverflow.com/questions/32127524/"
                        "how-to-install-and-use-make-in-windows")
            raise ValueError(msg)

        super().build(installation_dir=installation_dir)
        make_path = os.path.join(installation_dir, f"Makefile.{platform_name}")
        if not os.path.exists(make_path):
            make_path = os.path.join(installation_dir, f"Makefile")
        if not os.path.exists(make_path) and installation_dir == "darwin":
            make_path = os.path.join(installation_dir, f"Makefile.linux")
            if os.path.exists(make_path):
                enb.logger.warn("Warning! Neither a specific nor a generic Makefile is available "
                                "for this plugin and your platform. However, a Makefile for linux "
                                "is. The plugin will be attempted to be installed using that "
                                "Makefile.")
        if os.path.exists(make_path):
            invocation = (f"cd {os.path.dirname(os.path.abspath(make_path))} "
                          f"&& {shutil.which('make')} -f {os.path.basename(make_path)}")
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise Exception(f"Error building plugin {repr(cls.name)} with "
                                f"the Makefile at {make_path} ({status=})!\n"
                                f"Failing command:\n{textwrap.indent(invocation, ' ' * 3)}\n"
                                f"Output:\n{textwrap.indent(output, ' ' * 3)}")
        else:
            raise ValueError(f"Cannot build {repr(cls.name)}: no valid makefile "
                             f"in {installation_dir}.")


class PluginJava(Plugin):
    @classmethod
    def build(cls, installation_dir):
        if shutil.which("java") is None:
            enb.logger.warn(f"Warning! The 'java' program was not found in the path, "
                            f"but is required by the {repr(cls.name)} plugin. "
                            f"Installing anyway...")
        super().build(installation_dir=installation_dir)



#!/usr/bin/env python3
"""FITS format manipulation tools.
See https://fits.gsfc.nasa.gov/fits_documentation.html.
"""
__author__ = "Òscar Maireles"
__since__ = "2020/04/01"

# pylint: disable=no-self-use

import os
import tempfile

import numpy as np
from astropy.io import fits

from enb.atable import get_all_input_files, redefines_column
from enb.sets import FileVersionTable, FilePropertiesTable
from enb.isets import load_array_bsq, dump_array_bsq
from enb.compression.wrapper import WrapperCodec


class FITSVersionTable(FileVersionTable, FilePropertiesTable):
    """Read FITS files and convert them to raw files, sorting them by type (
    integer or float) and by bits per pixel.
    """
    fits_extension = "fit"
    allowed_extensions = ["fit", "fits"]
    version_name = "FitsToRaw"

    # No need to set  dataset_files_extension here,
    # because get_default_target_indices is overwritten.
    def __init__(self, original_base_dir, version_base_dir):
        """:param version_base_dir: path to the versioned base directory
          (versioned directories preserve names and structure within
          the base dir)

        :param original_base_dir: path to the original directory
          (it must contain all indices requested later with self.get_df()).
          If None, options.base_datset_dir is used
        """
        super().__init__(
            original_base_dir=original_base_dir,
            version_base_dir=version_base_dir,
            original_properties_table=sets.FilePropertiesTable(),
            version_name=self.version_name,
            check_generated_files=False)

    def get_default_target_indices(self):
        indices = []
        for ext in self.allowed_extensions:
            indices.extend(get_all_input_files(
                ext=ext, base_dataset_dir=self.original_base_dir))
        return indices

    def original_to_versioned_path(self, original_path):
        if original_path.lower().endswith(".fit"):
            input_ext = "fit"
        elif original_path.lower().endswith(".fits"):
            input_ext = "fits"
        else:
            raise ValueError(f"Invalid input extension {original_path}")

        return os.path.join(
            os.path.dirname(
                os.path.abspath(original_path)).replace(
                os.path.abspath(self.original_base_dir),
                os.path.abspath(self.version_base_dir)),
            os.path.basename(original_path).replace(
                f".{input_ext}", ".raw"))

    @redefines_column
    def set_version_repetitions(self, file_path, row):
        """Set the number of times the versioning process is performed.
        """
        # pylint: disable=unused-argument
        row[_column_name] = 1

    def version(self, input_path, output_path, row):
        # pylint: disable=too-many-branches,too-many-statements,too-many-locals
        if not input_path.lower().endswith(".fit") \
                and not input_path.lower().endswith(".fits"):
            raise ValueError(f"Invalid extension found in {input_path}")

        hdul = fits.open(input_path, ignore_missing_simple=True)
        saved_images = 0
        for hdu_index, hdu in enumerate(hdul):
            if hdu.header["NAXIS"] == 0:
                continue
            data = hdu.data.transpose()
            header = hdu.header

            if header['BITPIX'] == 8:
                pass
            else:

                if header['NAXIS'] == 1:
                    if header['BITPIX'] < 0:
                        name_label = f'-f{-header["BITPIX"]}-1x1x{header["NAXIS1"]}'
                        dtype_name = f'float{-header["BITPIX"]}'
                        enb_type_name = f"f{-header['BITPIX']}"
                    elif header['BITPIX'] > 0:
                        name_label = f'-u{header["BITPIX"]}be-1x1x{header["NAXIS1"]}'
                        dtype_name = f'>u{header["BITPIX"] // 8}'
                        enb_type_name = f"u{header['BITPIX']}be"
                    else:
                        raise ValueError(f"Invalid bitpix {header['BITPIX']}")
                    data = np.expand_dims(data, axis=1)
                    data = np.expand_dims(data, axis=2)

                elif header['NAXIS'] == 2:
                    if header['BITPIX'] < 0:
                        name_label = f'-f{-header["BITPIX"]}' \
                                     f'-1x{header["NAXIS2"]}x{header["NAXIS1"]}'
                        dtype_name = f'float{-header["BITPIX"]}'
                        enb_type_name = f"f{-header['BITPIX']}"
                    elif header['BITPIX'] > 0:
                        name_label = f'-u{header["BITPIX"]}be' \
                                     f'-1x{header["NAXIS2"]}x{header["NAXIS1"]}'
                        dtype_name = f'>u{header["BITPIX"] // 8}'
                        enb_type_name = f"u{header['BITPIX']}be"
                    else:
                        raise ValueError(f"Invalid bitpix {header['BITPIX']}")

                    data = np.expand_dims(data, axis=2)
                elif header['NAXIS'] == 3:
                    if header['BITPIX'] < 0:
                        name_label = f'-f{-header["BITPIX"]}' \
                                     f'-{header["NAXIS3"]}' \
                                     f'x{header["NAXIS2"]}' \
                                     f'x{header["NAXIS1"]}'
                        dtype_name = f'float{-header["BITPIX"]}'
                        enb_type_name = f"f{-header['BITPIX']}"
                    elif header['BITPIX'] > 0:
                        name_label = f'-u{header["BITPIX"]}be-' \
                                     f'{header["NAXIS3"]}' \
                                     f'x{header["NAXIS2"]}' \
                                     f'x{header["NAXIS1"]}'
                        dtype_name = f'>u{header["BITPIX"] // 8}'
                        enb_type_name = f"u{header['BITPIX']}be"
                    else:
                        raise ValueError(f"Invalid bitpix {header['BITPIX']}")
                elif header['NAXIS'] == 4:
                    if header['BITPIX'] < 0:
                        name_label = f'-f{-header["BITPIX"]}' \
                                     f'-{header["NAXIS3"]}' \
                                     f'x{header["NAXIS2"]}' \
                                     f'x{header["NAXIS1"]}'
                        dtype_name = f'float{-header["BITPIX"]}'
                        enb_type_name = f"f{-header['BITPIX']}"
                    elif header['BITPIX'] > 0:
                        name_label = f'-u{header["BITPIX"]}be' \
                                     f'-{header["NAXIS3"]}' \
                                     f'x{header["NAXIS2"]}' \
                                     f'x{header["NAXIS1"]}'
                        dtype_name = f'>u{header["BITPIX"] // 8}'
                        enb_type_name = f"u{header['BITPIX']}be"
                    else:
                        raise ValueError(f"Invalid bitpix {header['BITPIX']}")
                    data = np.squeeze(data, axis=3)
                else:
                    raise Exception(
                        f"Invalid header['NAXIS'] = {header['NAXIS']}")

                output_dir = os.path.join(
                    os.path.dirname(os.path.abspath(output_path)),
                    enb_type_name)
                effective_output_path = os.path.join(
                    output_dir,
                    f"{os.path.basename(output_path).replace('.raw', '')}"
                    f"_img{saved_images}{name_label}.raw")
                os.makedirs(os.path.dirname(effective_output_path),
                            exist_ok=True)
                if os.path.isfile(effective_output_path):
                    pass
                else:
                    if options.verbose > 2:
                        print(
                            f"Dumping FITS->raw ({repr(effective_output_path)})"
                            f" from hdu_index={hdu_index}")
                    dump_array_bsq(array=data, file_or_path=effective_output_path, dtype=dtype_name)
                    fits_header_path = os.path.join(os.path.dirname(
                        os.path.abspath(effective_output_path)).replace(
                        os.path.abspath(self.version_base_dir),
                        f"{os.path.abspath(self.version_base_dir)}_headers"),
                        os.path.basename(effective_output_path).replace(
                            '.raw', '') + "-fits_header.txt")
                    os.makedirs(os.path.dirname(fits_header_path),
                                exist_ok=True)
                    if options.verbose > 2:
                        print(
                            f"Writing to fits_header_path={repr(fits_header_path)}")
                    if os.path.exists(fits_header_path):
                        os.remove(fits_header_path)
                    header.totextfile(fits_header_path)

            saved_images += 1


class FITSWrapperCodec(WrapperCodec):
    """Raw images are coded into FITS before compression with the wrapper,
    and FITS is decoded to raw after decompression.
    """

    # pylint: disable=abstract-method

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        img = load_array_bsq(
            file_or_path=original_path, image_properties_row=original_file_info)
        with tempfile.NamedTemporaryFile(suffix=".fits") as tmp_file:
            os.remove(tmp_file.name)
            array = img.swapaxes(0, 2)
            hdu = fits.PrimaryHDU(array)
            hdu.writeto(tmp_file.name)

            if os.path.exists(compressed_path):
                os.remove(compressed_path)
            compression_results = super().compress(original_path=tmp_file.name,
                                                   compressed_path=compressed_path,
                                                   original_file_info=original_file_info)
            crs = self.compression_results_from_paths(
                original_path=original_path, compressed_path=compressed_path)
            crs.compression_time_seconds = max(0,
                                               compression_results.compression_time_seconds)
            crs.maximum_memory_kb = compression_results.maximum_memory_kb
            return crs

    def decompress(self, compressed_path, reconstructed_path,
                   original_file_info=None):

        with tempfile.NamedTemporaryFile(suffix=".fits") as tmp_file:
            os.remove(tmp_file.name)
            decompression_results = super().decompress(
                compressed_path=compressed_path,
                reconstructed_path=tmp_file.name)

            hdul = fits.open(tmp_file.name)
            assert len(hdul) == 1
            data = hdul[0].data.transpose()
            header = hdul[0].header

            if header['NAXIS'] == 2:
                if header['BITPIX'] < 0:
                    dtype_name = f'float{-header["BITPIX"]}'
                elif header['BITPIX'] > 0:
                    dtype_name = f'>u{header["BITPIX"] // 8}'
                else:
                    raise ValueError(f"Invalid bitpix {header['BITPIX']}")
            elif header['NAXIS'] == 3:
                if header['BITPIX'] < 0:
                    dtype_name = f'float{-header["BITPIX"]}'
                elif header['BITPIX'] > 0:
                    dtype_name = f'>u{header["BITPIX"] // 8}'
                else:
                    raise ValueError(f"Invalid bitpix {header['BITPIX']}")
            else:
                raise Exception(f"Invalid header['NAXIS'] = {header['NAXIS']}")

            dump_array_bsq(array=data,
                           file_or_path=reconstructed_path,
                           dtype=dtype_name)
            decompression_results.compressed_path = compressed_path
            decompression_results.reconstructed_path = reconstructed_path
            return decompression_results



#!/usr/bin/env python3
"""Unit tests for isets.py
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/04/07"

import unittest
import tempfile
import numpy as np

import test_all
from enb import pgm


def get_test_component(bytes_per_sample, width=67, height=34):
    array = np.zeros((width, height), dtype=f"u{bytes_per_sample}")
    max_pixel_value = 2 ** (8 * bytes_per_sample) - 1
    i = 0
    for y in range(height):
        for x in range(width):
            array[x, y] = (i % (max_pixel_value + 1))
            i += 1
    array[0, 0] = max_pixel_value
    assert array[0, 0] == max_pixel_value
    return array


class TestPGM(unittest.TestCase):
    def test_pgm_read_write(self):
        for bytes_per_sample in [1, 2]:
            array = get_test_component(bytes_per_sample=bytes_per_sample)

            with tempfile.NamedTemporaryFile() as tmp_file:
                pgm.write_pgm(array=array, bytes_per_sample=bytes_per_sample, output_path=tmp_file.name)
                loaded_array = pgm.read_pgm(input_path=tmp_file.name)
                assert array.shape == loaded_array.shape
                assert (array == loaded_array).all(), (array, '\n', loaded_array)

    def test_ppm_read_write(self):
        for bytes_per_sample in [1]:
            array = np.dstack([get_test_component(bytes_per_sample=bytes_per_sample)] * 3)
            with tempfile.NamedTemporaryFile() as tmp_file:
                pgm.write_ppm(array=array, bytes_per_sample=bytes_per_sample, output_path=tmp_file.name)
                loaded_array = pgm.read_ppm(input_path=tmp_file.name)
                assert array.shape == loaded_array.shape, (array.shape, loaded_array.shape)
                assert (array == loaded_array).all(), (array[:3,:3,:], loaded_array[:3,:3,:])


if __name__ == '__main__':
    unittest.main()



import textwrap
import enb.plugins


class VVCPlugin(enb.plugins.PluginMake):
    name = "vvc"
    label = "Wrapper for the VVC / H.266 codec"
    tags = {"data compression", "image", "codec"}
    contrib_authors = ["ISO", "ITU", "IEC"]
    contrib_reference_urls = ["https://vcgit.hhi.fraunhofer.de/jvet/VVCSoftware_VTM/-/tree/master"]
    contrib_download_url_name = [
        ("https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/"
         "VVCSoftware_VTM-VTM-15.0.zip?raw=true",
         "VVCSoftware_VTM-master.zip")]
    extra_requirements_message = textwrap.dedent("""
    The cmake tool is needed to build this plugin, if not present already. 
    You can install cmake as follows: 
    
    * Ubuntu: `sudo apt install cmake`
    * MacOS:
        1. Install homebrew from http://brew.sh
        2. `brew install gcc@9`
    """)
    tested_on = {"linux"}



import enb.plugins

class ZstandardPlugin(enb.plugins.PluginMake):
    name = "zstd"
    label = "Zstandard library wrapper"
    tags = {"data compression", "codec"}
    authors = ["Òscar Maireles"]
    contrib_authors = ["Yann Collet"]
    contrib_reference_urls = ["https://github.com/facebook/zstd"]
    contrib_download_url_name = [
        ("https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/zstd-1.1.3.zip?raw=true",
         "zstd-1.1.3.zip")]
    tested_on = {"linux"}



#!/usr/bin/env python3
"""Generate the sample vectors for the codec availability test.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/02/24"

import os
import shutil
import numpy as np
import enb
from enb.config import options

def generate_test_images(base_dir=os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")):
    width = 128
    height = 128

    for bytes_per_sample in [1, 2, 4]:
        enb.logger.verbose(f"\tgenerating vectors with {bytes_per_sample} "
                           f"byte{'s' if bytes_per_sample != 1 else ''} per sample...")

        for label, component_count in [("mono", 1), ("rgb", 3), ("rgba", 4), ("multi", 32)]:
            for signed in [True, False]:
                for big_endian in [True, False]:
                    for is_float in [True, False]:
                        if (is_float and bytes_per_sample < 2) \
                                or (is_float and not signed) \
                                or (is_float and not big_endian) \
                                or (signed and bytes_per_sample < 2) \
                                or (not big_endian and bytes_per_sample <= 1):
                            continue

                        if is_float:
                            # Float  samples
                            effective_bits_per_sample = 8 * bytes_per_sample

                            output_dir = os.path.join(
                                base_dir,
                                f"{label}_float",
                                f"{label}_f{effective_bits_per_sample}")

                            if os.path.isdir(output_dir) and not options.force:
                                continue

                            shutil.rmtree(output_dir, ignore_errors=True)
                            os.makedirs(output_dir)

                            type = f"f{effective_bits_per_sample}"
                            dtype = f"f{bytes_per_sample}"
                            geometry = f"{component_count}x{height}x{width}"
                            output_path = os.path.join(output_dir, f"testsample_{type}-{geometry}.raw")
                            total_samples = width * height * component_count

                            if bytes_per_sample == 2:
                                min_sample_value = np.finfo(np.float16).min
                                max_sample_value = np.finfo(np.float16).max
                            elif bytes_per_sample == 4:
                                min_sample_value = np.finfo(np.float32).min
                                max_sample_value = np.finfo(np.float32).max

                            samples = max_sample_value * np.random.rand(total_samples).astype(np.float64)
                            samples[::2] *= -1
                            samples[0] = min_sample_value
                            samples[1] = max_sample_value
                            with open(output_path, "wb") as output_file:
                                output_file.write(bytes(samples.astype(dtype)))

                        else:
                            # Integer samples
                            type = f"{'s' if signed else 'u'}{8 * bytes_per_sample}{'be' if big_endian else 'le'}"
                            dtype = f"{'>' if big_endian else '<'}{'i' if signed else 'u'}{bytes_per_sample}"
                            geometry = f"{component_count}x{height}x{width}"
                            output_dir = os.path.join(
                                base_dir, f"{label}_{type}")

                            if os.path.isdir(output_dir) and not options.force:
                                continue

                            shutil.rmtree(output_dir, ignore_errors=True)
                            os.makedirs(output_dir)

                            for missing_bits in range(8 * bytes_per_sample):
                                effective_bits_per_sample = 8 * bytes_per_sample - missing_bits
                                output_path = os.path.join(output_dir, f"testsample_effective_dr_{effective_bits_per_sample}bit_{type}-{geometry}.raw")

                                total_samples = width * height * component_count
                                if signed:
                                    min_sample_value = - (2 ** (effective_bits_per_sample - 1))
                                    max_sample_value = 2 ** (effective_bits_per_sample - 1) - 1
                                else:
                                    min_sample_value = 0
                                    max_sample_value = 2 ** effective_bits_per_sample - 1
                                samples = np.zeros(total_samples)
                                for i in range(total_samples):
                                    samples[i] = min_sample_value + i % (max_sample_value - min_sample_value + 1)
                                samples[0] = min_sample_value
                                samples[1] = max_sample_value
                                with open(output_path, "wb") as output_file:
                                    output_file.write(bytes(samples.astype(dtype)))

if __name__ == '__main__':
    generate_test_images()



#!/usr/bin/env python3
"""Codec wrapper for the fpzip lossless image coder
"""
__author__ = "Òscar Maireles and Miguel Hernández-Cabronero"
__since__ = "2021/06/01"

import os
import enb


class Fpzip(enb.icompression.LosslessCodec, enb.icompression.NearLosslessCodec, enb.icompression.WrapperCodec):
    """Wrapper for the fpzip codec
    Allowed data type to be compressed: float 32
    """

    def __init__(self, fpzip_binary=os.path.join(os.path.dirname(__file__), "fpzip")):
        super().__init__(compressor_path=fpzip_binary,
                         decompressor_path=fpzip_binary,
                         param_dict=dict())

    @property
    def label(self):
        return "FPZIP"

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        assert original_file_info["bytes_per_sample"] == 4 and original_file_info[
            "float"] == True, 'data type must be float 32'
        dimensions = 1

        return f"-i {os.path.abspath(original_path)} " \
               f" -o {os.path.abspath(compressed_path)} " \
               f"-{dimensions} {original_file_info.samples} "

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        dimensions = 1

        return f"-d -i {compressed_path} -o {reconstructed_path} -{dimensions} {original_file_info.samples} "



#!/usr/bin/env python3
"""FSE's Huff0 codec wrapper.
"""
__author__ = "Òscar Maireles and Miguel Hernández-Cabronero"
__since__ = "2021/06/01"

import os
from enb import icompression


class FSEHuffman(icompression.WrapperCodec, icompression.LosslessCodec):
    """FSE's Huff0 codec wrapper (block adaptive)
    """
    def __init__(self, huff_binary=os.path.join(os.path.dirname(os.path.abspath(__file__)), "fse")):
        icompression.WrapperCodec.__init__(self,
                                           compressor_path=huff_binary,
                                           decompressor_path=huff_binary,
                                           param_dict=None, output_invocation_dir=None)

        self.huff_binary = huff_binary

        assert os.path.isfile(huff_binary)

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        try:
            os.remove(compressed_path)
        except FileNotFoundError:
            pass
        return f"-h {original_path} {compressed_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        try:
            os.remove(reconstructed_path)
        except FileNotFoundError:
            pass
        return f"-h -d {compressed_path} {reconstructed_path}"

    @property
    def label(self):
        return "Huffman (block adaptive)"



#!/usr/bin/env python3
"""Codec wrapper for the montsec software
"""
__author__ = "Xavier Fernandez Mellado and Pau Quintas Torra"
__since__ = "2023/12/14"

import os

import enb


class Montsec(enb.icompression.LosslessCodec, enb.icompression.NearLosslessCodec,
              enb.icompression.JavaWrapperCodec, enb.icompression.GiciLibHelper):
    """Wrapper for the montsec codec
    """

    def __init__(self,
                 compressor_jar=os.path.join(os.path.dirname(__file__), "montsec.jar"),
                 decompressor_jar=os.path.join(os.path.dirname(__file__), "montsec.jar"),
                 qs: int = 1, cm: int = 0, bypass: int = 0):
        """
        :param qs: sets the quantization step.
        :param cm: context model
            0.- No context model is used.
            1.- Context modelling type 1 is used during the encoding process (self context).
            2.- Context modelling type 2 is used during the encoding process (HV).
            3.- Context modelling type 3 is used during the encoding process (HVDS).
        :param bypass: Number of the bitplane in which stop compressing and send the raw bit
        """
        assert (cm >= 0) and (cm <= 3), \
            f"The context modeling must be a integer value between [0 and 3]"
        assert (bypass >= 0) and (bypass <= 8), \
            f"Bypass must be into [0 and 8]"
        super().__init__(compressor_jar=compressor_jar,
                         decompressor_jar=decompressor_jar,
                         param_dict=dict(qs=qs, cm=cm, bypass=bypass))

    @property
    def label(self):
        return ("Montsec"
                + (" self context" if self.param_dict['cm'] == 1 else "")
                + (" HV context" if self.param_dict['cm'] == 2 else "")
                + (" HVDS context" if self.param_dict['cm'] == 3 else ""))

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        assert not original_file_info["float"], \
            f"Only integer samples are currently supported by {self.__class__.__name__}"
        assert original_file_info["bytes_per_sample"] in [1, 2], \
            f"Bytes per sample must be 1 or 2 in class {self.__class__.__name__}"
        assert original_file_info["big_endian"], \
            f"Only big-endian samples are currently supported by {self.__class__.__name__}"

        return f"-Xmx24g -jar {self.compressor_jar} -c -i {original_path} -o {compressed_path} " \
               f"-ig {self.get_gici_geometry_str(original_file_info=original_file_info)} " \
               f"-qs {self.param_dict['qs']} " \
               f"-cm {self.param_dict['cm']} " \
               f"-bypass {self.param_dict['bypass']} "

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        assert not original_file_info["float"], \
            f"Only integer samples are currently supported by {self.__class__.__name__}"
        assert original_file_info["bytes_per_sample"] in [1, 2], \
            f"Bytes per sample must be 1 or 2 in class {self.__class__.__name__}"
        assert original_file_info["big_endian"], \
            f"Only big-endian samples are currently supported by {self.__class__.__name__}"

        return f"-Xmx24g -jar {self.decompressor_jar} -d -i {compressed_path} -o {reconstructed_path} "



#!/usr/bin/env python3
"""Test the codec.py module
"""
import unittest
import os
import filecmp
import shutil
import random

random.seed(0xbadc0f33)

from test_all import options
from codec_implementations import trivial_codecs as trivial_codecs


class TestAllCodecs(unittest.TestCase):
    """Test the AbstractCodec class
    """

    def test_trivial_lossless(self):
        """Test the trivial raw-output codec.
        """
        for cls in [trivial_codecs.TrivialLosslessCodec,
                    trivial_codecs.TrivialCpWrapper]:
            test_one_codec(cls())


def test_one_codec(codec_instance):

    if options.verbose:
        print(f"\n\tLossless check for {codec_instance.name}")

    FL, packet_count = 3, 1024
    original_path = f"test_raw.FL{FL}bytes.raw"
    compressed_path = "compressed.bin"
    reconstructed_path = "reconstructed.raw"
    try:
        output_bytes = bytearray()
        for _ in range(FL * packet_count):
            output_bytes.append(random.randint(0, 255))
        output_file = open(original_path, "wb")
        output_file.write(output_bytes)
        output_file.close()
        cr = codec_instance.compress(original_path=original_path,
                                     compressed_path=compressed_path)
        dr = codec_instance.decompress(compressed_path=compressed_path,
                                       reconstructed_path=reconstructed_path)
        if cr:
            assert cr.original_path == original_path
            assert cr.compressed_path == compressed_path
            assert cr.codec_name == codec_instance.name
        if dr:
            assert dr.compressed_path == compressed_path
            assert dr.reconstructed_path == reconstructed_path
        if cr and dr:
            assert dr.codec_name == cr.codec_name
            assert dr.codec_param_dict == cr.codec_param_dict
        assert os.path.exists(reconstructed_path), \
            f"Codec {codec_instance.name} did not produce any compressed file"
        assert filecmp.cmp(original_path, reconstructed_path), \
            (codec_instance, original_path, compressed_path, reconstructed_path)
    finally:
        for p in (original_path, compressed_path, reconstructed_path):
            if os.path.exists(p):
                os.remove(p)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""Codecs wrapping V2F forests.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/09/01"

import os
import shutil
import subprocess
import itertools
import re
import pandas as pd
import numpy as np
import enb
import time

# Constants taken from the C code for consistency and maintainability.
V2F_C_QUANTIZER_MODE_NONE = 0
V2F_C_QUANTIZER_MODE_UNIFORM = 1
V2F_C_QUANTIZER_MODE_COUNT = 2
V2F_C_DECORRELATOR_MODE_NONE = 0
V2F_C_DECORRELATOR_MODE_LEFT = 1
V2F_C_DECORRELATOR_MODE_2_LEFT = 2
V2F_C_DECORRELATOR_MODE_JPEGLS = 3
V2F_C_DECORRELATOR_MODE_FGIJ = 4
V2F_C_DECORRELATOR_MODE_COUNT = 5
V2F_C_QUANTIZER_MODE_MAX_STEP_SIZE = 255


class V2FCodec(enb.icompression.LosslessCodec,
               enb.icompression.NearLosslessCodec,
               enb.icompression.WrapperCodec):
    """Wrapper for the V2F compressor/decompressor tools in C,
    which must be present in the same path as this script.
    """

    def __init__(self, v2fc_header_path, qstep=None, quantizer_mode=None, decorrelator_mode=None,
                 verify_on_initialization=True, time_results_dir=None,
                 shadow_position_pairs=[]):
        """
        Initialize a V2F codec that employs the forest definition in `v2fc_header_path`, and optionally
        overwrite the quantization and decorrelation defined in them.

        :param v2fc_header_path: path to the V2F forest definition to be employed
        :param qstep: quantization step size. Only applied if quantizer_mode is not
          V2F_C_DECORRELATOR_MODE_NONE. Cannot exceed V2F_C_QUANTIZER_MODE_MAX_STEP_SIZE.
        :param quantizer_mode: quantization mode; see the V2F_C_QUANTIZER_MODE_* constants.
        :param decorrelator_mode: decorrelation (e.g., prediction) mode;
          see see the V2F_C_DECORRELATOR_MODE_* constants.
        :param verify_on_initialization: if True, the codec in v2fc_heder_path is verified upon initialization.
        :param time_results_dir: if not None, it must be the path of a directory that will be used to store time
          measurements reported by the V2F prototype.
        :param shadow_position_pairs: a list of zero or more tuples describing the y-positions of the 
          horizontal shadow regions. Shadow regions are not compressed and are reconstructed as constant zero.
          Each tuple must be of the form (start, end), where start and end are the first and last row index 
          (starting with zero) of the shadow. If the list is empty, no shadow regions are specified.
          Shadow regions cannot overlap. 
        """
        self.codec_root_dir = os.path.abspath(os.path.dirname(__file__))

        shadow_position_pairs = sorted(shadow_position_pairs)
        verify_shadow_position_pairs(shadow_position_pairs)

        super().__init__(compressor_path=os.path.join(self.codec_root_dir, "v2f_compress"),
                         decompressor_path=os.path.join(self.codec_root_dir, "v2f_decompress"),
                         param_dict=dict(
                             v2fc_header_path=v2fc_header_path,
                             qstep=qstep,
                             quantizer_mode=quantizer_mode,
                             decorrelator_mode=decorrelator_mode,
                             shadow_position_pairs=shadow_position_pairs))

        # Header path with the V2F codec (including forest) definition
        self.v2fc_header_path = v2fc_header_path
        if not self.v2fc_header_path or not os.path.isfile(self.v2fc_header_path):
            raise ValueError(f"Cannot find a V2F codec definition at {repr(self.v2fc_header_path)}. "
                             f"Please make sure you are passing the right argument to "
                             f"{self.__class__}'s initializier, and that the file is actually present.")

        # Verify the chosen codec if configured to do so.
        self.verifier_path = os.path.join(self.codec_root_dir, "v2f_verify_codec")
        if verify_on_initialization:
            self.verify_codec(v2fc_header_path=v2fc_header_path)

        # Prepare the folder where time measurements are to be stored
        self.time_results_dir = time_results_dir
        if self.time_results_dir:
            os.makedirs(self.time_results_dir, exist_ok=True)

    def verify_codec(self, v2fc_header_path):
        """Run the V2F codec verification tool for a given path.
        An exception is raised if this verification fails.
        """
        if not os.path.isfile(self.verifier_path):
            raise ValueError(f"Could not find the V2F codec verifier tool at {repr(self.verifier_path)}. "
                             f"Please make sure it is present at that path and try again.")

        with enb.logger.info_context(f"Verifying V2F codec {repr(v2fc_header_path)}"):
            invocation = f"{self.verifier_path} {os.path.abspath(v2fc_header_path)}"
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise ValueError(f"V2F codec verification failed for {repr(v2fc_header_path)}.\n"
                                 f"status: {status}\n"
                                 f"invocation: {repr(invocation)}\n"
                                 f"execution output: {output}.")
            else:
                enb.logger.info(f"Verification OK. Returned message:\n{output}")

    def get_time_path(self, original_path):
        assert self.time_results_dir, f"Trying to run get_time_path() without having set self.time_results_dir."
        return os.path.join(self.time_results_dir,
                            f"q{self.param_dict['quantizer_mode']}_"
                            f"s{self.param_dict['qstep']}_"
                            f"d{self.param_dict['decorrelator_mode']}_" + \
                            ("noshadow_" if self.param_dict["shadow_position_pairs"] else "") + \
                            os.path.basename(self.v2fc_header_path).replace(os.sep, "__"),
                            os.path.relpath(os.path.abspath(original_path), 
                                            enb.config.options.project_root).replace(os.sep, "__")
                            + "_times.csv")

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        return f"{original_path} {self.get_optional_argument_string(original_path=original_path)} " \
               f"-w {original_file_info['width']} " \
               + (f"-y {','.join(str(y) for y in itertools.chain(*self.param_dict['shadow_position_pairs']))} "
                  if self.param_dict['shadow_position_pairs'] else '') + \
               f"{self.v2fc_header_path} {compressed_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"{compressed_path} " \
               f"{self.get_optional_argument_string(original_path=None)} " \
               f"-w {original_file_info['width']} " \
               f"{self.v2fc_header_path} {reconstructed_path}"

    def get_optional_argument_string(self, original_path):
        optional_str = ""
        if self.param_dict['quantizer_mode'] is not None:
            optional_str += f"-q {self.param_dict['quantizer_mode']} "
        if self.param_dict['qstep'] is not None:
            optional_str += f"-s {self.param_dict['qstep']} "
        if self.param_dict['decorrelator_mode'] is not None:
            optional_str += f"-d {self.param_dict['decorrelator_mode']} "
        if self.time_results_dir is not None and original_path is not None:
            time_path = self.get_time_path(original_path)
            os.makedirs(os.path.dirname(time_path), exist_ok=True)
            optional_str += f"-t {time_path} "
        return optional_str

    @property
    def label(self):
        if self.param_dict["decorrelator_mode"] == V2F_C_DECORRELATOR_MODE_NONE:
            dec_str = "none"
        elif self.param_dict["decorrelator_mode"] == V2F_C_DECORRELATOR_MODE_LEFT:
            dec_str = "left"
        elif self.param_dict["decorrelator_mode"] == V2F_C_DECORRELATOR_MODE_2_LEFT:
            dec_str = "two left"
        elif self.param_dict["decorrelator_mode"] == V2F_C_DECORRELATOR_MODE_JPEGLS:
            dec_str = "JPEG-LS"
        elif self.param_dict["decorrelator_mode"] == V2F_C_DECORRELATOR_MODE_FGIJ:
            dec_str = "FGIJ"
        else:
            raise ValueError(self.param_dict["decorrelator_mode"])

        return f"V2F Qstep={self.param_dict['qstep']} pred={dec_str}"


class ShadowLossyExperiment(enb.icompression.LossyCompressionExperiment):
    """This compression experiment verifies that loss is restricted to 
    the defined y shadow regions, while the remainder of the image is
    reconstructed without loss. An exception is raised if this verification fails.
    
    Some additional columns are generated to track tree size, decorrelation mode, and
    other parameters specific to V2F codecs.
    """

    def __init__(self, codecs, shadow_position_pairs=[],
                 dataset_paths=None,
                 csv_experiment_path=None,
                 csv_dataset_path=None,
                 dataset_info_table=None,
                 overwrite_file_properties=False,
                 reconstructed_dir_path=None,
                 compressed_copy_dir_path=None,
                 task_families=None):
        super().__init__(codecs=codecs,
                         dataset_paths=dataset_paths,
                         csv_experiment_path=csv_experiment_path,
                         csv_dataset_path=csv_dataset_path,
                         dataset_info_table=dataset_info_table,
                         overwrite_file_properties=overwrite_file_properties,
                         reconstructed_dir_path=reconstructed_dir_path,
                         compressed_copy_dir_path=compressed_copy_dir_path,
                         task_families=task_families)
        verify_shadow_position_pairs(shadow_position_pairs)
        self.shadow_position_pairs = sorted(shadow_position_pairs)

    def column_lossless_except_shadows(self, index, row):
        """Verify that lossless results are obtained in every part of the image except for
        the shadows, which must contain all zero values if the shadow_position_pairs entry
        is present in the codec's param_dict.
        """
        original_path, codec = self.index_to_path_task(index)
        reconstructed_path = self.codec_results.decompression_results.reconstructed_path

        original_img = enb.isets.load_array_bsq(original_path)
        reconstructed_img = enb.isets.load_array_bsq(
            file_or_path=reconstructed_path,
            image_properties_row=self.get_dataset_info_row(file_path=original_path))

        for shadow_start, shadow_end in self.shadow_position_pairs:
            try:
                if codec.param_dict["shadow_position_pairs"]:
                    if np.any(reconstructed_img[:, shadow_start:shadow_end + 1, :] != 0):
                        raise enb.icompression.CompressionException(
                            output=f"Non-zero values found at ({shadow_start}, {shadow_end}) of the reconstructed image")
            except KeyError:
                pass
            reconstructed_img[:, shadow_start:shadow_end + 1, :] = \
                original_img[:, shadow_start:shadow_end + 1, :]

        if np.any(original_img != reconstructed_img):
            raise enb.icompression.CompressionException(
                output=f"Not lossless reconstruction outside the shadow regions")

        return True

    def column_decorrelator_mode(self, index, row):
        """Name of the decorrelation method used by this current codec.
        """
        path, codec = self.index_to_path_task(index)

        try:
            decorrelator_mode = codec.param_dict["decorrelator_mode"]
            if decorrelator_mode == V2F_C_DECORRELATOR_MODE_NONE:
                return "No decorrelation"
            elif decorrelator_mode == V2F_C_DECORRELATOR_MODE_LEFT:
                return "Left neighbor prediction"
            elif decorrelator_mode == V2F_C_DECORRELATOR_MODE_2_LEFT:
                return "Two left neighbor prediction"
            elif decorrelator_mode == V2F_C_DECORRELATOR_MODE_JPEGLS:
                return "JPEG-LS prediction"
            elif decorrelator_mode == V2F_C_DECORRELATOR_MODE_FGIJ:
                return "Four neighbor prediction"
            else:
                raise KeyError(decorrelator_mode)
        except KeyError:
            assert not isinstance(codec, V2FCodec)
            return f"{codec.label}'s decorrelation"

    def column_tree_size(self, index, row):
        """Number of included nodes in the trees of the V2F forest.
        """
        path, codec = self.index_to_path_task(index)
        try:
            return int(re.search(r"_treesize-(\d+)_", os.path.basename(codec.v2fc_header_path)).group(1))
        except AttributeError:
            return -1

    def column_tree_count(self, index, row):
        """Number of trees in the V2F forest.
        """
        path, codec = self.index_to_path_task(index)
        try:
            return int(re.search(r"_treecount-(\d+)_", os.path.basename(codec.v2fc_header_path)).group(1))
        except AttributeError:
            return -1

    def column_symbol_count(self, index, row):
        """Number of symbols in the source.
        """
        path, codec = self.index_to_path_task(index)
        try:
            return int(re.search(r"_symbolcount-(\d+)_", os.path.basename(codec.v2fc_header_path)).group(1))
        except AttributeError:
            return -1

    def column_optimization_column(self, index, row):
        """Name of the analysis column used to obtain the ideal source's
        symbol distribution.
        """
        path, codec = self.index_to_path_task(index)
        try:
            return str(re.search(r"_optimizedfor-(.*)_withentropy-", os.path.basename(codec.v2fc_header_path)).group(1))
        except AttributeError:
            return f"{codec.label}: N/A"

    def column_ideal_source_entropy(self, index, row):
        """Entropy of the ideal source for which the forest is optimized.
        """
        path, codec = self.index_to_path_task(index)
        try:
            return float(
                re.search(r"_withentropy-(\d+\.\d+)_avg_all.v2fc", os.path.basename(codec.v2fc_header_path)).group(1))
        except AttributeError:
            return -1

    @enb.atable.column_function("block_coding_time_seconds", label="Block coding time (s)", plot_min=0)
    def set_block_coding_time_seconds(self, index, row):
        """Calculate the total coding time without considering initialization,
        """
        path, codec = self.index_to_path_task(index)
        if not isinstance(codec, V2FCodec):
            row[_column_name] = row["compression_time_seconds"]
        else:
            time_df = pd.read_csv(codec.get_time_path(path))
            row[_column_name] = float(time_df[time_df["name"] == "v2f_compressor_compress_block"]["total_cpu_seconds"])


def verify_shadow_position_pairs(shadow_position_pairs):
    """Check that the list of shadow positions is valid, and raise a ValueError otherwise.
    """
    if shadow_position_pairs:
        shadow_position_pairs = sorted(shadow_position_pairs)
        if any(len(pair) != 2 for pair in shadow_position_pairs):
            print(f"[watch] shadow_position_pairs={shadow_position_pairs}")

            raise ValueError("All shadow pairs must have length exactly 2")
        if any(pair[0] > pair[1] for pair in shadow_position_pairs):
            raise ValueError("Shadow pairs must have start <= end")
        for i in range(len(shadow_position_pairs) - 1):
            if shadow_position_pairs[i][1] >= shadow_position_pairs[i + 1][0]:
                raise ValueError(f"Shadow region {shadow_position_pairs[i]} and {shadow_position_pairs[i + 1]} "
                                 "overlap, which is not allowed.")



#!/usr/bin/env python3
"""Installation script for the enb library.

Adapted from
https://www.jeffknupp.com/blog/2013/08/16/open-sourcing-a-python-project-the-right-way/.

Refer to the user manual (https://miguelinux314.github.io/experiment-notebook)
for additional information on how to install this software.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2019/09/19"

import os
import importlib
import configparser
from setuptools import setup, find_packages

setup_package_list = ["setuptools", "wheel"]

for module_name in setup_package_list:
    try:
        importlib.import_module(module_name)
    except (ModuleNotFoundError, ImportError) as ex:
        raise ModuleNotFoundError(
            f"\n\n{'@' * 80}\n"
            f"{'@' * 80}\n"
            "\n"
            f"Package {module_name} needs to be installed in your python environment "
            f"to be able to install enb.\n"
            f"The full list of pre-installation requirements is: "
            f"{', '.join(setup_package_list)}.\n\n"
            f"Please run `pip install {' '.join(setup_package_list)}` "
            f"before installing enb\n\n"
            f"{'@' * 80}\n"
            f"{'@' * 80}\n"
            "\n") from ex

# Read the configuration from ./enb/config/enb.ini, section "enb"
enb_options = configparser.ConfigParser()
enb_options.read(os.path.join(os.path.dirname(os.path.abspath(__file__)), "enb", "config", "enb.ini"))
enb_options = enb_options["enb"]

with open("README.md", "r") as readme_file:
    setup(
        # Metadata about the project
        name=enb_options["name"],
        version=enb_options["version"],
        url=enb_options["url"],
        download_url=enb_options["download_url"],
        license=enb_options["license"],
        author=enb_options["author"],
        author_email=enb_options["author_email"],
        description=enb_options["description"],
        long_description=readme_file.read(),
        long_description_content_type="text/markdown",
        platforms=enb_options["platforms"],
        python_requires=enb_options["python_requires"],
        classifiers=[
            "Programming Language :: Python",
            f"Development Status :: {enb_options['development_status']}",
            "Natural Language :: English",
            "Environment :: Console",
            "Intended Audience :: Developers",
            f"License :: OSI Approved :: {enb_options['license']}",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Topic :: Scientific/Engineering",
        ],

        # UI
        entry_points={
            # Main CLI entry point
            "console_scripts": ["enb=enb.__main__:main"]
        },

        # Dependencies
        setup_requires=setup_package_list,

        install_requires=[
            "appdirs",
            "astropy",
            "deprecation",
            "h5py",
            "imageio",
            "jinja2>=3.1.2",
            "matplotlib",
            "natsort",
            "numpngw",
            "numpy",
            "pandas>=2.1.0",
            "pathos",
            "pdf2image",
            "psutil",
            "requests",
            "rich",
            "scipy",
            "sortedcontainers",
        ],

        # This part determines the contents of the installed folder in your python's
        # site-packages location.
        # MANIFEST.in is assumed to have been updated, i.e., via git hooks.
        # This allows core plugins and templates to be automatically included.
        packages=[p for p in find_packages() if p.startswith("enb")],
        include_package_data=True)



"""Wrapper codec classes. 

Existing codec implementations (including non-python binaries) can be easily added to enb
via WrapperCodec (sub)classes. 
"""
import ast
import os
import functools
import shutil
import tempfile
import time
import json
import numpy as np

import enb
from enb import logger, tcall
from enb import isets
from enb.config import options
from enb.atable import get_canonical_path
from enb.compression import CompressionResults, DecompressionResults, CompressionException, DecompressionException
from enb.compression.codec import AbstractCodec, NearLosslessCodec
from enb.compression import tarlite


class WrapperCodec(AbstractCodec):
    """A codec that uses an external process to compress and decompress.
    """

    def __init__(self, compressor_path, decompressor_path, param_dict=None,
                 output_invocation_dir=None,
                 signature_in_name=False):
        """
        :param compressor_path: path to the executable to be used for
          compression
        :param decompressor_path: path to the executable to be used for
          decompression
        :param param_dict: name-value mapping of the parameters to be used
          for compression
        :param output_invocation_dir: if not None, invocation strings are
          stored in this directory with name based on the codec and the sample's
          full path.
        :pram signature_in_name: if True, the default codec name includes
          part of the hexdigest of the compressor and decompressor binaries
          being used
        """
        # Set relative paths so that local and remote workers can use the same command line
        if os.path.abspath(compressor_path).startswith(os.path.abspath(options.project_root)):
            compressor_path = get_canonical_path(compressor_path)
        if os.path.abspath(decompressor_path).startswith(os.path.abspath(options.project_root)):
            decompressor_path = get_canonical_path(decompressor_path)

        # pylint: disable=too-many-arguments
        super().__init__(param_dict=param_dict)
        self.signature_in_name = False
        if os.path.isfile(compressor_path):
            self.compressor_path = compressor_path
        else:
            self.compressor_path = shutil.which(compressor_path)
        if os.path.exists(decompressor_path):
            self.decompressor_path = decompressor_path
        else:
            self.decompressor_path = shutil.which(decompressor_path)

        if not self.compressor_path or not os.path.isfile(self.compressor_path):
            raise FileNotFoundError(
                f"Compressor path {repr(compressor_path)} is not available "
                f"for {self.__class__.__name__}")
        if not self.decompressor_path or not os.path.isfile(
                self.decompressor_path):
            raise FileNotFoundError(
                f"Decompressor path {repr(compressor_path)} is not available "
                f"for {self.__class__.__name__}")
        self.output_invocation_dir = output_invocation_dir

    def get_compression_params(self, original_path, compressed_path,
                               original_file_info):
        """Return a string (shell style) with the parameters to be passed to
        the compressor.

        Same parameter semantics as :meth:`AbstractCodec.compress`.

        :param original_file_info: a dict-like object describing
          original_path's properties (e.g., geometry), or None
        """
        raise NotImplementedError()

    def get_decompression_params(self, compressed_path, reconstructed_path,
                                 original_file_info):
        """Return a string (shell style) with the parameters to be passed to
        the decompressor. Same parameter semantics as
        :meth:`AbstractCodec.decompress()`.

        :param original_file_info: a dict-like object describing
          original_path's properties (e.g., geometry), or None
        """
        raise NotImplementedError()

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        compression_params = self.get_compression_params(
            original_path=original_path,
            compressed_path=compressed_path,
            original_file_info=original_file_info)
        invocation = f"{self.compressor_path} {compression_params}"
        logger.debug(f"[{self.name}] Invocation: '{invocation}'")
        try:
            logger.debug(f"[{self.name}] executing: {repr(invocation)}")
            status, output, measured_time, memory_kb = \
                tcall.get_status_output_time_memory(invocation=invocation)
            logger.debug(
                f"[{self.name}] Compression OK; "
                f"invocation={invocation} - status={status}; "
                f"output={output}; memory={memory_kb} KB")
        except tcall.InvocationError as ex:
            raise CompressionException(
                original_path=original_path,
                compressed_path=compressed_path,
                file_info=original_file_info,
                status=-1,
                output=None) from ex

        compression_results = self.compression_results_from_paths(
            original_path=original_path, compressed_path=compressed_path)
        compression_results.compression_time_seconds = measured_time
        compression_results.maximum_memory_kb = memory_kb

        if self.output_invocation_dir is not None:
            os.makedirs(self.output_invocation_dir, exist_ok=True)
            invocation_name = "invocation_compression_" \
                              + self.name \
                              + os.path.abspath(
                os.path.realpath(original_file_info["file_path"])).replace(
                os.sep, "_")
            with open(os.path.join(self.output_invocation_dir, invocation_name),
                      "w") as invocation_file:
                invocation_file.write(
                    f"Original path: {original_path}\n"
                    f"Compressed path: {compressed_path}\n"
                    f"Codec: {self.name}\n"
                    f"Invocation: {invocation}\n"
                    f"Status: {status}\n"
                    f"Output: {output}\n"
                    f"Measured time: {measured_time} s\n"
                    f"Maximum resident memory: {memory_kb} kb")

        return compression_results

    def decompress(self, compressed_path, reconstructed_path,
                   original_file_info=None):
        decompression_params = self.get_decompression_params(
            compressed_path=compressed_path,
            reconstructed_path=reconstructed_path,
            original_file_info=original_file_info)
        invocation = f"{self.decompressor_path} {decompression_params}"
        logger.debug(
            f"WrapperCodec({self.__class__.__name__}:decompress invocation={invocation}")
        try:
            status, output, measured_time, memory_kb \
                = tcall.get_status_output_time_memory(invocation)
            logger.debug(
                f"[{self.name}] Compression OK; "
                f"invocation={invocation} - status={status}; "
                f"output={output}; memory={memory_kb} kb")
        except tcall.InvocationError as ex:
            raise DecompressionException(
                compressed_path=compressed_path,
                reconstructed_path=reconstructed_path,
                file_info=original_file_info,
                status=-1,
                output=None) from ex

        decompression_results = self.decompression_results_from_paths(
            compressed_path=compressed_path,
            reconstructed_path=reconstructed_path)
        decompression_results.decompression_time_seconds = measured_time
        decompression_results.maximum_memory_kb = memory_kb

        if self.output_invocation_dir is not None:
            invocation_name = "invocation_decompression_" \
                              + self.name \
                              + os.path.abspath(os.path.realpath(
                original_file_info["file_path"] \
                    if original_file_info is not None \
                    else compressed_path)).replace(os.sep, "_")
            with open(os.path.join(self.output_invocation_dir, invocation_name),
                      "w") as invocation_file:
                invocation_file.write(
                    f"Compressed path: {compressed_path}\n"
                    f"Reconstructed path: {reconstructed_path}\n"
                    f"Codec: {self.name}\n"
                    f"Invocation: {invocation}\n"
                    f"Status: {status}\n"
                    f"Output: {output}\n"
                    f"Measured time: {measured_time} s\n"
                    f"Maximum resident size: {memory_kb} kb")

        return decompression_results

    @staticmethod
    @functools.lru_cache(maxsize=2)
    def get_binary_signature(binary_path):
        """Return a string with a (hopefully) unique signature for the
        contents of binary_path. By default, the first 5 digits of the
        sha-256 hexdigest are returned.
        """
        hasher = hashlib.sha256()
        with open(binary_path, "rb") as open_file:
            hasher.update(open_file.read())
        return hasher.hexdigest()[:5]

    @property
    def name(self):
        """Return the codec's name and parameters, also including the encoder
        and decoder hash summaries (so that changes in the reference binaries
        can be easily detected)
        """
        signature = None
        if self.signature_in_name:
            compressor_signature = self.get_binary_signature(
                self.compressor_path)
            decompressor_signature = self.get_binary_signature(
                self.decompressor_path)
            if compressor_signature and decompressor_signature:
                if compressor_signature == decompressor_signature:
                    signature = f"{compressor_signature}"
                else:
                    signature = f"{compressor_signature}_{compressor_signature}"

        name = f"{self.__class__.__name__}_{signature}" \
            if signature is not None else self.__class__.__name__
        if self.param_dict:
            name += "__" + "_".join(
                f"{k}={v}" for k, v in sorted(self.param_dict.items()))
        return name


class QuantizationWrapperCodec(NearLosslessCodec):
    """Perform uniform scalar quantization before compressing and after decompressing with
    a wrapped codec instance. Midpoint reconstruction is used in the dequantization stage.
    """

    def __init__(self, codec: AbstractCodec, qstep: int):
        """
        :param codec: The codec instance used to compress and decompress the quantized data.
        :param qstep: The quantization interval length
        """
        if qstep < 1 or int(qstep) != qstep:
            raise ValueError("The quantization step must be an integer no smaller than 1")
        super().__init__(param_dict=dict(codec_name=codec.name, qstep=qstep))
        self.codec = codec

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        with tempfile.TemporaryDirectory(dir=options.base_tmp_dir) as tmp_dir:
            # Read original data
            array = isets.load_array_bsq(original_path)

            # Apply quantization
            if not (self.param_dict["qstep"] & (self.param_dict["qstep"] - 1)):
                # Efficiently way to execute image_array // self.param_dict["qstep"]
                # When qstep is a power of 2
                array >>= self.param_dict["qstep"].bit_length() - 1
            else:
                array //= self.param_dict["qstep"]

            # Apply compression to the quantized data
            output_quantized_path = os.path.join(tmp_dir, os.path.basename(original_path))
            isets.dump_array_bsq(array, output_quantized_path)
            self.codec.compress(output_quantized_path, compressed_path, original_file_info)

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        with tempfile.TemporaryDirectory(dir=options.base_tmp_dir) as tmp_dir:
            output_quantized_path = os.path.join(tmp_dir, os.path.basename(reconstructed_path))
            self.codec.decompress(compressed_path, output_quantized_path, original_file_info)
            array = isets.load_array_bsq(output_quantized_path)

            if not (self.param_dict["qstep"] & (self.param_dict["qstep"] - 1)):
                # Efficiently way to execute image_array * self.param_dict["qstep"]
                # When qstep is a power of 2
                array <<= self.param_dict["qstep"].bit_length() - 1
            else:
                array *= self.param_dict["qstep"]

            min_val, max_val = np.iinfo(array.dtype).min, np.iinfo(array.dtype).max
            array = np.clip((array.astype(np.int64) + (self.param_dict["qstep"] // 2)),
                            min_val, max_val).astype(array.dtype)
            isets.dump_array_bsq(array, reconstructed_path)

    @property
    def name(self):
        """Return the original codec name and the quantization parameter
        """
        return f"{self.codec.name}_qstep={self.param_dict['qstep']}"

    @property
    def label(self):
        """Return the original codec label and the quantization parameter.
        """
        return f"{self.codec.label}, Q$_\\mathrm{{step}}=${self.param_dict['qstep']}"


class JavaWrapperCodec(WrapperCodec):
    """Wrapper for `*.jar` codecs. The compression and decompression
    parameters are those that need to be passed to the 'java' command.

    The `compressor_jar` and `decompressor_jar` attributes are added upon
    initialization based on the params to `__init__`.
    """

    def __init__(self, compressor_jar, decompressor_jar, param_dict=None):
        assert shutil.which("java") is not None, \
            f"The 'java' program was not found in the path, but is required by " \
            f"{self.__class__.__name__}. " \
            f"Please (re)install a JRE in the path and try again."
        super().__init__(compressor_path=shutil.which("java"),
                         decompressor_path=shutil.which("java"),
                         param_dict=param_dict)
        self.compressor_jar = get_canonical_path(compressor_jar)
        self.decompressor_jar = get_canonical_path(decompressor_jar)


class LittleEndianWrapper(WrapperCodec):
    """Wrapper with identical semantics as WrapperCodec, but performs a big
    endian to little endian conversion for (big-endian) 2-byte and 4-byte
    samples. If the input is flagged as little endian, e.g., if -u16le- is in
    the original file name, then no transformation is performed.

    Codecs inheriting from this class automatically receive little-endian
    samples, and are expected to reconstruct little-endian files (which are
    then translated back to big endian if and only if the original image was
    flagged as big endian.
    """

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        if original_file_info["big_endian"] and original_file_info["bytes_per_sample"] > 1:
            with tempfile.NamedTemporaryFile(
                    dir=options.base_tmp_dir,
                    suffix=f"-u{8 * original_file_info['bytes_per_sample']}le"
                           f"-{original_file_info['component_count']}"
                           f"x{original_file_info['height']}"
                           f"x{original_file_info['width']}.raw") \
                    as reversed_endian_file:
                be_img = isets.load_array_bsq(
                    file_or_path=original_path,
                    image_properties_row=original_file_info)
                reversed_file_info = dict(original_file_info)
                reversed_file_info["big_endian"] = False
                sign_str = "i" if original_file_info["signed"] else "u"
                isets.dump_array_bsq(array=be_img.astype(
                    f"<{sign_str}{original_file_info['bytes_per_sample']}"),
                    file_or_path=reversed_endian_file.name)
                compression_results = super().compress(
                    original_path=reversed_endian_file.name,
                    compressed_path=compressed_path,
                    original_file_info=reversed_file_info)
                compression_results.original_path = original_path
                return compression_results
        else:
            return super().compress(
                original_path=original_path,
                compressed_path=compressed_path,
                original_file_info=original_file_info)

    def decompress(self, compressed_path, reconstructed_path,
                   original_file_info=None):
        if original_file_info["big_endian"] and original_file_info["bytes_per_sample"] > 1:
            with tempfile.NamedTemporaryFile(
                    dir=options.base_tmp_dir,
                    suffix=f"-u{8 * original_file_info['bytes_per_sample']}le"
                           f"-{original_file_info['component_count']}"
                           f"x{original_file_info['height']}"
                           f"x{original_file_info['width']}.raw") as reversed_endian_file:
                reversed_file_info = dict(original_file_info)
                reversed_file_info["big_endian"] = True
                decompression_results = super().decompress(
                    compressed_path=compressed_path,
                    reconstructed_path=reversed_endian_file.name,
                    original_file_info=reversed_file_info)
                le_img = isets.load_array_bsq(
                    file_or_path=reversed_endian_file.name,
                    image_properties_row=reversed_file_info)
                sign_str = "i" if original_file_info["signed"] else "u"
                # Store with < (le) instead of > (be) to undo the byte reversal during compression
                isets.dump_array_bsq(array=le_img.astype(
                    f"<{sign_str}{original_file_info['bytes_per_sample']}"),
                    file_or_path=reconstructed_path)
                decompression_results.reconstructed_path = reconstructed_path

                return decompression_results
        else:
            return super().decompress(compressed_path=compressed_path,
                                      reconstructed_path=reconstructed_path,
                                      original_file_info=original_file_info)


class ReindexWrapper(AbstractCodec):
    """Input samples are first reindexed to a contiguous support preserving the ordering 
    (If x and y are two sample values present in the input file, then
    x < y <=> reindex(x) < reindex(y)).
    
    Reindexed data are stored as unsigned, big-endian samples of the width configured
    on initialization. After reindexing, the codec passed to the initializer is used for compression.
    
    The user is responsible for using a codec compatible with the type of the reindexed data,
    and a data type that can hold the number of unique samples present in the input file.
    
    Note that only integer input samples are currently supported.
    """

    def __init__(self, codec: AbstractCodec, width_bytes: int):
        super().__init__(param_dict={"width_bytes": width_bytes})
        self.codec = codec

    def compress(self, original_path: str, compressed_path: str, original_file_info=None) -> CompressionResults:
        with tempfile.TemporaryDirectory(dir=options.base_tmp_dir) as tmp_dir:
            # Find the unique values and store them as side information
            array = enb.isets.load_array_bsq(original_path)
            time_before = time.process_time() if not enb.config.options.report_wall_time else time.time()
            unique_values = np.unique(array)
            if len(unique_values) > 2 ** (8 * self.param_dict['width_bytes']):
                raise ValueError(
                    f"Trying to reindex into a {self.param_dict['width_bytes']}-byte file "
                    f"but there are {len(unique_values)} (too many) unique values.")
            side_info_path = os.path.join(tmp_dir, "side_info")
            with open(side_info_path, "w") as side_info_file:
                side_info_file.write(json.dumps(
                    obj=dict(original_dtype=str(array.dtype),
                             unique_values=str([int(x) for x in unique_values]))))

            # Reindex keeping the value sorting
            reindex_array = np.zeros(shape=array.shape, dtype=f">u{self.param_dict['width_bytes']}")
            for i, val in enumerate(unique_values):
                reindex_array[array == val] = i

            time_after = time.process_time() if not enb.config.options.report_wall_time else time.time()
            reindex_time = time_after - time_before

            # Store the reindexed data
            reindexed_path = os.path.join(tmp_dir, f"reindexed_{os.path.basename(original_path)}")
            for tag in enb.isets.dtype_tags:
                reindexed_path = os.path.join(
                    os.path.dirname(reindexed_path),
                    os.path.basename(reindexed_path).replace(tag, f"u{8 * self.param_dict['width_bytes']}be"))
            enb.isets.dump_array_bsq(array=reindex_array, file_or_path=reindexed_path)

            # Compress the reindexed data
            reindexed_compressed_path = os.path.join(tmp_dir, f"data")
            time_before = time.process_time() if not enb.config.options.report_wall_time else time.time()
            reindex_file_info = dict(original_file_info) if original_file_info is not None else None
            if reindex_file_info is not None:
                reindex_file_info["signed"] = False
                reindex_file_info["big_endian"] = True
                reindex_file_info["bytes_per_sample"] = 2
            compression_results = self.codec.compress(reindexed_path, reindexed_compressed_path, reindex_file_info)
            time_after = time.process_time() if not enb.config.options.report_wall_time else time.time()
            compression_time = time_after - time_before

            # Update the reported compression results
            if compression_results is None:
                compression_results = self.codec.compression_results_from_paths(
                    original_path=reindexed_path,
                    compressed_path=compressed_path)
            compression_results.original_path = original_path
            compression_results.compressed_path = compressed_path
            if compression_results.compression_time_seconds is None \
                    or compression_results.compression_time_seconds < 0:
                compression_results.compression_time_seconds = compression_time
            compression_results.compression_time_seconds += reindex_time

            # Store the side information and reindexed data into a tarlite package
            tarlite.tarlite_files([side_info_path, reindexed_compressed_path], compressed_path)

            return compression_results

    def decompress(self, compressed_path: str, reconstructed_path: str,
                   original_file_info=None) -> DecompressionResults:
        with tempfile.TemporaryDirectory(dir=options.base_tmp_dir) as tmp_dir:
            # Extract the tarlite package
            tarlite.untarlite_files(input_tarlite_path=compressed_path, output_dir_path=tmp_dir)
            reindexed_compressed_path = os.path.join(tmp_dir, "data")
            side_info_path = os.path.join(tmp_dir, "side_info")

            # Read and complete the side information
            with open(side_info_path, "r") as side_info_file:
                side_info = json.load(side_info_file)
                unique_values = np.array(ast.literal_eval(side_info["unique_values"]))
                original_width_bytes = int(side_info["original_dtype"][-1])

            # Decompress the reindexed data
            reindexed_path = os.path.join(tmp_dir, f"reindexed")
            self.codec.decompress(reindexed_compressed_path, reindexed_path, original_file_info)

            # Invert the reindexing
            assert os.path.getsize(reindexed_path) % original_width_bytes == 0, \
                ("Invalid output file size", os.path.getsize(reindexed_path), original_width_bytes)

            reindexed_data = enb.isets.load_array(
                file_or_path=reindexed_path,
                width=os.path.getsize(reindexed_path) // self.param_dict['width_bytes'],
                height=1,
                component_count=1,
                dtype=f">u{self.param_dict['width_bytes']}")

            data = np.zeros(shape=reindexed_data.shape, dtype=np.int64)
            np.take(unique_values, reindexed_data, out=data)
            enb.isets.dump_array_bsq(array=data.astype(side_info["original_dtype"]), file_or_path=reconstructed_path)

    @property
    def name(self):
        """Return the original codec name and the quantization parameter
        """
        return f"{self.codec.name}_reindex{self.param_dict['width_bytes']}B"

    @property
    def label(self):
        """Return the original codec label and the quantization parameter.
        """
        return f"Reindex {self.param_dict['width_bytes']}B + {self.codec.label}"



#!/usr/bin/env python3
"""Tools to run compression experiments
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2019/09/19"

import os
import collections
import itertools
import inspect
import time
import datetime

import enb.atable
import enb.sets
from enb.config import options


class ExperimentTask:
    """Identify an :py:class:`Experiment`'s task and its configuration.
    """

    def __init__(self, param_dict=None):
        """
        :param param_dict: dictionary of configuration parameters used
          for this task. By default, they are part of the task's name.
        """
        self.param_dict = dict(param_dict) if param_dict is not None else {}

    def apply(self, experiment, index, row):
        """This method is called for all combinations of input indices and tasks
        before computing any additional column.

        :param experiment: experiment invoking this method
        :param index: index being processed by the experiment. Consider
          experiment.index_to_path_task(index).
        :param row: row being currently processed by the experiment
        """

    @property
    def name(self):
        """Unique name that uniquely identifies a task and its configuration.
        It is not intended to be displayed as much as to be used as an index.
        """
        name = f"{self.__class__.__name__}"
        if self.param_dict:
            name += "__" + "_".join(
                f"{k}={v}" for k, v in sorted(self.param_dict.items()))
        return name

    @property
    def label(self):
        """Label to be displayed for the task in output reports.
        May not be strictly unique nor fully informative.
        By default, the task's name is returned.
        """
        return self.name

    def __eq__(self, other):
        try:
            # pylint: disable=unidiomatic-typecheck
            return all(v == other.param_dict[k]
                       for k, v in self.param_dict.items()) \
                and all(self.param_dict[k] == v
                        for k, v in other.param_dict.items()) \
                and type(self) == type(other)
        except (KeyError, AttributeError):
            return False

    def __hash__(self):
        return tuple(sorted(self.param_dict.items())).__hash__()

    def __repr__(self):
        return f"<{self.__class__.__name__} (" \
            + ', '.join(repr(param) + '=' + repr(value)
                        for param, value in self.param_dict.items()) \
            + ")>"


class Experiment(enb.atable.ATable):
    """An Experiment allows seamless execution of one or more tasks upon a
    corpus of test files. Tasks are identified by a file index and a
    :py:class:`ExperimentTask`'s (unique) name.

    For each task, any number of table columns can be defined to gather
    results of interest. This allows easy extension to highly complex and/or
    specific experiments. See :func:`set_task_name` for an example.

    Automatic persistence of the obtained results is provided, to allow faster
    experiment development and result replication.

    Internally, an Experiment consists of an ATable containing the properties
    of the test corpus, and is itself another ATable that contains the
    results of an experiment with as many user-defined columns as needed.
    When the :meth:`~Experiment.get_df` method is called, a joint DataFrame
    instance is returned that contains both the experiment results, and the
    associated metainformation columns for each corpus element.
    """
    task_name_column = "task_name"
    task_label_column = "task_label"
    task_apply_time_column = "task_apply_time"
    default_file_properties_table_class = enb.sets.FilePropertiesTable
    no_family_label = "No family"

    def __init__(self, tasks,
                 dataset_paths=None,
                 csv_experiment_path=None,
                 csv_dataset_path=None,
                 dataset_info_table=None,
                 overwrite_file_properties=False,
                 task_families=None):
        """
        :param tasks: an iterable of :py:class:`ExperimentTask` instances.
          Each test file is processed by all defined tasks. For each (file,
          task) combination, a row is included in the table returned by
          :py:meth:`~get_df()`.
        :param dataset_paths: list of paths to the files to be used as input.
          If it is None, this list is obtained automatically calling
          `enb.sets.get_all_test_file()`
        :param csv_experiment_path: if not None, path to the CSV file giving
          persistence support to this experiment. If None, it is automatically
          determined within options.persistence_dir.
        :param csv_dataset_path: if not None, path to the CSV file given
          persistence support to the dataset file properties. If None,
          it is automatically determined within options.persistence_dir.
        :param dataset_info_table: if not None, it must be a
          enb.sets.FilePropertiesTable instance or subclass instance that can
          be used to obtain dataset file metainformation, and/or gather it from
          csv_dataset_path. If None, a new enb.sets.FilePropertiesTable
          instance is created and used for this purpose. This parameter can
          also be a class (instead of an instance). In this case, it the
          initializer is asumed to accept a csv_support_path argument and be
          compatible with the enb.sets.FilePropertiesTable interface.
        :param overwrite_file_properties: if True, file properties are
          necessarily computed before starting the experiment. This can be
          useful for temporary and/or random datasets. If False,
          file properties are loaded from the persistent storage when
          available. Note that this parameter does not affect whether
          experiment results are retrieved from persistent storage if
          available. This is controlled via the parameters passed to get_df()
        :param task_families: if not None, it must be a list of TaskFamily
          instances. It is used to set the "family_label" column for each row.
          If the codec is not found within the families, a default label is set
          indicating so.
        """
        # pylint: disable=too-many-arguments
        overwrite_file_properties = overwrite_file_properties \
            if overwrite_file_properties is not None else options.force

        self.tasks = list(tasks)
        self.tasks_by_name = collections.OrderedDict(
            {str(task.name): task for task in self.tasks})
        self.task_families = task_families
        self.task_name_to_family_label = {}
        try:
            for task_family in task_families:
                for task_name in task_family.task_names:
                    if task_name in self.task_name_to_family_label:
                        raise ValueError(
                            f"Found task_name {repr(task_name)} "
                            f"in family {task_family.label} "
                            "that was already present in family "
                            f"{self.task_name_to_family_label[task_name]}. "
                            "No duplicates are allowed.")
                    assert task_name not in self.task_name_to_family_label, \
                        self.task_name_to_family_label[task_name]
                    self.task_name_to_family_label[
                        task_name] = task_family.label
        except TypeError:
            pass

        if dataset_paths is not None:
            dataset_paths = [enb.atable.get_canonical_path(p) for p in
                             dataset_paths]
        else:
            dataset_paths = enb.atable.get_all_input_files(
                ext=self.dataset_files_extension)

        if csv_dataset_path is None:
            csv_dataset_path = os.path.join(
                options.persistence_dir,
                f"{self.__class__.__name__}_dataset_persistence.csv")
        os.makedirs(os.path.dirname(csv_dataset_path), exist_ok=True)

        if dataset_info_table is None:
            dataset_info_table = self.default_file_properties_table_class(
                csv_support_path=csv_dataset_path)
        else:
            if inspect.isclass(dataset_info_table):
                dataset_info_table = dataset_info_table(
                    csv_support_path=csv_dataset_path)
        self.dataset_info_table = dataset_info_table

        self.dataset_info_table.ignored_columns = \
            set(self.dataset_info_table.ignored_columns + self.ignored_columns)

        assert len(self.dataset_info_table.indices) == 1, \
            "dataset_info_table is expected to have a single index"

        enb.logger.debug(f"[{self.__class__.__name__}] ", end="")
        enb.logger.info(f"Obtaining dataset properties table for {self.__class__.__name__}...")
        self.dataset_table_df = self.dataset_info_table.get_df(
            target_indices=dataset_paths,
            overwrite=overwrite_file_properties,
            fill=True)

        self.target_file_paths = dataset_paths

        if csv_experiment_path is None:
            csv_experiment_path = os.path.join(
                options.persistence_dir,
                f"persistence_{self.__class__.__name__}.csv")

        os.makedirs(os.path.dirname(csv_experiment_path), exist_ok=True)

        super().__init__(csv_support_path=csv_experiment_path,
                         index=self.dataset_info_table.indices + [
                             self.task_name_column])

    def get_df(self, target_indices=None, target_columns=None,
               fill=True, overwrite=None, chunk_size=None):
        """Get a DataFrame with the results of the experiment. The produced
        DataFrame contains the columns from the dataset info table (but they
        are not stored in the experiment's persistence file).

        :param target_indices: list of file paths to be processed. If None,
          self.target_file_paths is used instead.
        :param chunk_size: if not None, a positive integer that determines
          the number of table rows that are processed before made persistent.
        :param overwrite: if not None, a flag determining whether existing
          values should be calculated again. If none, options
        """
        # pylint: disable=too-many-arguments,too-many-locals
        target_indices = self.target_file_paths \
            if target_indices is None else target_indices
        overwrite = overwrite if overwrite is not None else options.force

        if not self.tasks:
            raise ValueError(
                f"No tasks were defined for {self.__class__.__name__}.")

        target_task_names = [str(t.name) for t in self.tasks]
        target_indices = tuple(itertools.product(
            sorted(set(target_indices)), sorted(set(target_task_names))))

        enb.logger.info(f"Computing {self.__class__.__name__} ({len(self.column_to_properties)} columns)...")
        enb.logger.debug(f"({len(self.tasks)} tasks, {len(target_indices)} indices)")

        chunk_size = chunk_size if chunk_size is not None \
            else options.chunk_size
        chunk_size = chunk_size if chunk_size is not None \
            else len(target_indices)

        chunks = [target_indices[i:i + chunk_size]
                  for i in range(0, len(target_indices), chunk_size)]

        if enb.progress.is_progress_enabled():
            progress_tracker = enb.progress.ProgressTracker(atable=self,
                                                            row_count=len(target_indices),
                                                            chunk_size=chunk_size).__enter__()
        else:
            progress_tracker = False
        try:
            for chunk_index, chunk in enumerate(chunks):
                with enb.logger.debug_context(
                        f"Computing experiment chunk "
                        f"{chunk_index}/{len(chunks) - 1} "
                        f"({100 * chunk_index / len(chunks):.2f}%-"
                        f"{min(100, 100 * (chunk_index + 1) / len(chunks)):.2f}%) "
                        f"@ {datetime.datetime.now()}",
                        sep="...\n",
                        msg_after=f"Completed {self.__class__.__name__} "
                                  f"chunk #{chunk_index}/{len(chunks) - 1}"):
                    try:
                        old_tasks = list(self.tasks)
                        old_tasks_by_name = dict(self.tasks_by_name)
                        task_names = list(set(task_name for _, task_name in chunk))
                        self.tasks = [old_tasks_by_name[name] for name in task_names]
                        self.tasks_by_name = {name: old_tasks_by_name[name]
                                              for name in task_names}
                        _ = super().get_df(target_indices=chunk,
                                           fill=fill,
                                           overwrite=overwrite,
                                           chunk_size=-1,
                                           progress_tracker=progress_tracker)
                    finally:
                        self.tasks = old_tasks
                        self.tasks_by_name = old_tasks_by_name

                if progress_tracker:
                    progress_tracker.complete_chunk()
        finally:
            if progress_tracker:
                progress_tracker.__exit__(None, None, None)

        # Get all dfs (single chunk)
        df = super().get_df(target_indices=target_indices, fill=fill,
                            overwrite=overwrite,
                            chunk_size=len(target_indices),
                            progress_tracker=False)

        # Add dataset columns
        with enb.logger.debug_context(
                "Merging dataset and experiment results"):
            rsuffix = "__redundant__index"
            df = df.join(
                self.dataset_table_df.set_index(self.dataset_info_table.index),
                on=self.dataset_info_table.index, rsuffix=rsuffix)
            redundant_columns = [c.replace(rsuffix, "")
                                 for c in df.columns
                                 if c.endswith(rsuffix)
                                 and not c.startswith("row_created")
                                 and not c.startswith("row_updated")]
            if redundant_columns:
                enb.logger.warn(
                    f"Found redundant dataset/experiment column(s): "
                    f"{', '.join(redundant_columns)}.")

        enb.logger.info("")

        return df[(c for c in df.columns if not c.endswith(rsuffix))]

    def index_to_path_task(self, index):
        """Given an Experiment's row index, return `(path, task)`, where
        `path` is the canonical path of the row's dataset element,
        and `task` is the task instance corresponding to that row.
        """
        return index[0], self.tasks_by_name[index[1]]

    def get_dataset_info_row(self, file_path):
        """Get the dataset info table row for the file path given as argument.
        """
        return self.get_dataset_df().loc[
            enb.atable.indices_to_internal_loc(file_path)]

    def get_dataset_df(self):
        """Get the DataFrame of the employed dataset.
        """
        return self.dataset_table_df

    @property
    def joined_column_to_properties(self):
        """Get the combined dictionary of :class:`enb.atable.ColumnProperties`
        indexed by column name. This dictionary contains the dataset
        properties columns and the experiment columns.

        Note that :py:attr:`~enb.Experiment.column_to_properties` returns
        only the experiment columns.
        """
        property_dict = dict(self.dataset_info_table.column_to_properties)
        property_dict.update(self.column_to_properties)
        return property_dict

    @enb.atable.column_function(task_apply_time_column)
    def set_task_apply_time(self, index, row):
        """Run the `task.apply()` method and store its process time.
        """
        _, task = self.index_to_path_task(index)
        time_before = time.process_time()
        task.apply(self, index, row)
        row[_column_name] = max(0.0, time.process_time() - time_before)

    @enb.atable.column_function(task_name_column)
    def set_task_name(self, index, row):
        """Set the name of the task for this row.
        """
        _, task = self.index_to_path_task(index)
        row[_column_name] = task.name

    @enb.atable.column_function(task_label_column)
    def set_task_label(self, index, row):
        """Set the label of the task for this row.
        """
        _, task = self.index_to_path_task(index)
        row[_column_name] = task.label

    @enb.atable.column_function(
        enb.atable.ColumnProperties("param_dict", has_dict_values=True))
    def set_param_dict(self, index, row):
        """Store the task's param dict for easy reference and access to the
        param values.
        """
        _, task = self.index_to_path_task(index)
        row[_column_name] = task.param_dict

    @enb.atable.column_function("family_label")
    def set_family_label(self, index, row):
        """Set the label of the family to which this row's task belong,
        or set it to self.no_family_label
        """
        _, task = self.index_to_path_task(index)
        try:
            row[_column_name] = self.task_name_to_family_label[task.name]
        except KeyError:
            row[_column_name] = self.no_family_label


class TaskFamily:
    """Describe a sorted list of task names that identify a family of related
    results within a DataFrame. Typically, this family will be constructed
    using task workers (e.g., :class:`enb.icompression.AbstractCodec`
    instances) that share all configuration values except for a parameter.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, label, task_names=None, name_to_label=None):
        """
        :param label: Printable name that identifies the family
        :param task_names: if not None, it must be a list of task names (
          strings) that are expected to be found in an ATable's DataFrame when
          analyzing it.
        :param name_to_label: if not None, it must be a dictionary indexed by
          task name that contains a displayable version of it.
        """
        self.label = label
        self.task_names = task_names if task_names is not None else []
        self.name_to_label = name_to_label if name_to_label is not None else {}

    def add_task(self, task_name, task_label=None):
        """
        Add a new task name to the family (it becomes the last element in
        self.task_names)

        :param task_name: A new not previously included in the Family
        """
        # Any ExperimentTask-like object can also be passed, and its name is
        # automatically extracted
        try:
            task_name = task_name.name
        except (TypeError, AttributeError):
            pass

        assert task_name not in self.task_names
        self.task_names.append(task_name)
        if task_label:
            self.name_to_label[task_name] = task_label



import enb.plugins


class FileVersionExample(enb.plugins.Plugin):
    """Template and usage example for the FileVersionTable class.
    """
    name = "file_version_example"
    author = ["Miguel Hernández-Cabronero"]
    label = "Template and usage example for the FileVersionTable class for dataset curation"
    tags = {"project", "documentation"}
    tested_on = {"linux"}



#!/usr/bin/env python3
"""Unit tests for the aatable classes
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/05/07"

import os
import unittest
import pandas as pd
import tempfile
import enb


class TestScalarToScalar(unittest.TestCase):
    """Test the plotting of scalar to scalar dictionaries.
    """
    def test_scalar_to_scalar(self):
        df = pd.DataFrame(columns=["dicts"])
        for i in range(10):
            df.loc[len(df)] = pd.Series(dict(dicts=dict(a=i, b=10 * i, c=i ** 2 - 2 * i + 3)))
        for i in range(10):
            df.loc[len(df)] = pd.Series(dict(dicts=dict(a=i, b=10 * i)))
        for i in range(10):
            df.loc[len(df)] = pd.Series(dict(dicts=dict(a=i)))
        for i in range(10):
            df.loc[len(df)] = pd.Series(dict(dicts=dict(a=i, c=0)))

        with tempfile.TemporaryDirectory() as tmp_dir:
            analyzer = enb.aanalysis.DictNumericAnalyzer()
            analyzer.get_df(
                full_df=df,
                target_columns=["dicts"],
                column_to_properties={"dicts": enb.atable.ColumnProperties("dicts", has_dict_values=True)},
                output_plot_dir=tmp_dir)


if __name__ == "__main__":
    unittest.main()


import enb


class PNGToRawTemplate(enb.plugins.Template):
    name = "png-to-raw"
    author = ["Miguel Hernández-Cabronero"]
    label = "Curate a directory of PNG files into raw (fixed-length) format"
    tags = {"template", "data compression", "image"}
    tested_on = {"linux"}


#!/usr/bin/env python3
"""Example showing the basic workflow of the ``enb`` library in a simple ATable subclass.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/09/13"

import os
import glob
import shutil
import enb.atable
import enb.aanalysis


class WikiTable(enb.atable.ATable):
    # Methods that start with column_ are automatically recognized as such.
    # They just need to return the intended value.
    def column_line_count(self, file_path, row):
        with open(file_path, encoding="utf-8", mode="r") as input_file:
            return sum(1 for line in input_file)

    # More information about a column can be provided
    # Several column names and/or enb.atable.ColumnProperties instances
    # can also be passed as *args to the decorator to define multiple columns
    @enb.atable.column_function(
        enb.atable.ColumnProperties(name="word_count", label="Word count", plot_min=0),
        "status")
    def set_word_count(self, file_path, row):
        with open(file_path, mode="r", encoding="utf-8") as input_file:
            contents = input_file.read()
            row["word_count"] = len(contents.split())
            row["status"] = "dead" if "death_place" in contents.lower() else "alive"


def main():
    # Step 1: get a list of all data samples
    sample_paths = glob.glob("./data/wiki/*.txt")

    # Step 2: run experiment to gather data
    table = WikiTable() # csv_support_path="persistence_basic_workflow.csv")
    result_df = table.get_df(target_indices=sample_paths)

    # Show part of the returned table for the documentation.
    print(result_df[["index", "line_count"]])


    # Step 3: plot results
    #   Distribution of line counts
    analysis_df = enb.aanalysis.ScalarNumericAnalyzer().get_df(
        full_df=result_df,
        target_columns=["line_count"])
    os.makedirs("analysis", exist_ok=True)
    analysis_df.to_csv("analysis/line_count_analysis.csv")

    #   Distribution of word count grouped by status
    analysis_df = enb.aanalysis.ScalarNumericAnalyzer().get_df(
        full_df=result_df,
        target_columns=["word_count"],
        group_by="status",
        show_global=False)
    os.makedirs("analysis", exist_ok=True)
    analysis_df.to_csv("analysis/word_count_analysis.csv")

    #   Scatter plot: line count vs word count
    enb.aanalysis.TwoNumericAnalyzer().get_df(
        full_df=result_df,
        target_columns=[("line_count", "word_count")],
        column_to_properties=table.column_to_properties)


if __name__ == '__main__':
    main()



"""enb.compression: data compression in enb.

The compression and icompression modules implement enb.experiment.Experiment classes 
and other basic tools to facilitate them.

Several other modules are declared for specific compressed data formats.
"""

from .compression import *
from . import codec
from . import icompression
from . import wrapper

from . import tarlite
from . import fits
from . import jpg
from . import pgm
from . import png



#!/usr/bin/env python3
"""Wrappers for different types of LZ codecs such as Deflate, LZMA, BZIP2
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/01/04"

import zlib
import lzma
import bz2

from enb import icompression


class AbstractZipCodec(icompression.LosslessCodec):
    MIN_COMPRESSION_LEVEL = 1
    MAX_COMPRESSION_LEVEL = 9
    DEFAULT_COMPRESSION_LEVEL = 5

    def __init__(self, compression_level=DEFAULT_COMPRESSION_LEVEL, param_dict=None):
        assert self.MIN_COMPRESSION_LEVEL <= compression_level <= self.MAX_COMPRESSION_LEVEL
        param_dict = dict() if param_dict is None else param_dict
        param_dict["compression_level"] = compression_level
        super().__init__(param_dict=param_dict)


class LZ77Huffman(AbstractZipCodec):
    """Apply the LZ77 algorithm and Huffman coding to the file using zlib.
    """

    def compress(self, original_path, compressed_path, original_file_info):
        with open(original_path, "rb") as original_file, open(compressed_path, "wb") as compressed_file:
            compressed_file.write(zlib.compress(original_file.read(), level=self.param_dict["compression_level"]))

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        with open(compressed_path, "rb") as compressed_file, open(reconstructed_path, "wb") as reconstructed_file:
            reconstructed_file.write(zlib.decompress(compressed_file.read()))

    @property
    def label(self):
        return f"LZ77 Huffman level {self.param_dict['compression_level']}"


class LZMA(AbstractZipCodec):
    """Apply the LZMA algorithm using the lzma library
    """

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        with open(original_path, "rb") as input_file, open(compressed_path, "wb") as output_file:
            output_file.write(lzma.compress(input_file.read(), preset=self.param_dict["compression_level"]))

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        with open(compressed_path, "rb") as compressed_file, open(reconstructed_path, "wb") as reconstructed_file:
            reconstructed_file.write(lzma.decompress(compressed_file.read()))

    @property
    def label(self):
        return f"LZMA level {self.param_dict['compression_level']}"


class BZIP2(AbstractZipCodec):
    """Apply the BZIP2 algorithm using zlib.
    """

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        with open(original_path, "rb") as input_file, open(compressed_path, "wb") as output_file:
            output_file.write(bz2.compress(input_file.read(), self.param_dict["compression_level"]))

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        with open(compressed_path, "rb") as compressed_file, open(reconstructed_path, "wb") as reconstructed_file:
            reconstructed_file.write(bz2.decompress(compressed_file.read()))

    @property
    def label(self):
        return f"BZIP2 level {self.param_dict['compression_level']}"
